using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.UI;

namespace JayJayWon
{
    public class BrowserAssist : MVRScript
    {
        public static BrowserUI browserUI =null;

        SuperController.OnAtomUIDRename atomRenameHandle;
        SuperController.OnAtomRemoved atomRemoveHandle;
        SuperController.OnAtomAdded atomAddedHandle;
        SuperController.OnSceneLoaded sceneLoadedHandle;
        public override void Init()
        {
            try
            {
                Globals.mvrScript = this;
                if (containingAtom.name == "CoreControl" && containingAtom.type == "SessionPluginManager")
                {
                    atomRenameHandle = new SuperController.OnAtomUIDRename(this.OnAtomNameUpdate);
                    SuperController.singleton.onAtomUIDRenameHandlers += atomRenameHandle;

                    atomRemoveHandle = new SuperController.OnAtomRemoved(this.OnAtomRemoved);
                    SuperController.singleton.onAtomRemovedHandlers += atomRemoveHandle;

                    sceneLoadedHandle = new SuperController.OnSceneLoaded(OnSceneLoaded);
                    SuperController.singleton.onSceneLoadedHandlers += sceneLoadedHandle; 

                    atomAddedHandle = new SuperController.OnAtomAdded(OnAtomAdded);
                    SuperController.singleton.onAtomAddedHandlers += atomAddedHandle;

                    // Initiate integration with Acidbubbles KeyBindings
                    SuperController.singleton.BroadcastMessage("OnActionsProviderAvailable", this, SendMessageOptions.DontRequireReceiver);

                    Utils.CreatePluginDataFolder();
                    

                    DateTime scanTime = DateTime.Now;
                    VARPackageManifest.Init(scanTime);
                    ResourceManifest.LocalResourcesInit(scanTime);

                    browserUI = new BrowserUI();
                    browserUI.resourceBrowserUI.RefactorButtonRows((int)Settings.maxButtonsPerPageJSF.val, 6);

                    var button = CreateButton("Toggle UI On/Off");
                    button.button.onClick.AddListener(delegate
                    {
                        browserUI.ToggleActive();
                    });

                    var maxButtonsSlider = CreateSlider(Settings.maxButtonsPerPageJSF);
                    maxButtonsSlider.valueFormat = "F0";
                    maxButtonsSlider.ConfigureQuickButtons(-10f, -50f, -100f, -1000f, 10f, 50f, 100f, 1000f);

                    var uiScaleSlider = CreateSlider(Settings.uiScaleJSF);
                    uiScaleSlider.valueFormat = "F2";
                    uiScaleSlider.ConfigureQuickButtons(-0.1f, -0.25f, -0.5f, -1f, 0.1f, 0.25f, 0.5f, 1f);

                    Settings.LoadSettings();

                    CopySymLinkBatScripts();
                    CheckSymLinkAccess();
                }
                else
                {
                    JSONStorableString text = new JSONStorableString("text", "Browser Assist must be loaded as a Session Plugin");
                    CreateTextField(text);
                }

            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(Init)}: {e}");
                OnDestroy();
            }
        }

        private void CopySymLinkBatScripts()
        {
            try
            {
                string pluginSlot = name.Substring(0, name.IndexOf('_'));               
                JSONClass pluginsJC = manager.GetJSON(true, true, true).AsObject["plugins"].AsObject;

                string pluginPath = pluginsJC[pluginSlot];
                if (pluginPath.Contains(":"))
                {
                    string varName = pluginPath.Substring(0, pluginPath.IndexOf(':'));
                    string createSymLinksSourceFileName = FileManagerSecure.NormalizePath(varName + ":\\" + Const.BAScriptFolder + "\\CreateBASymLinks.bat");
                    string removeSymLinksSourceFileName = FileManagerSecure.NormalizePath(varName + ":\\" + Const.BAScriptFolder + "\\RemoveBASymLinks.bat");

                    if (FileManagerSecure.FileExists(createSymLinksSourceFileName))
                    {
                        FileManagerSecure.CopyFile(createSymLinksSourceFileName, Const.BAScriptDataFolder + "\\CreateBASymLinks.bat");
                    }
                    if (FileManagerSecure.FileExists(removeSymLinksSourceFileName))
                    {
                        FileManagerSecure.CopyFile(removeSymLinksSourceFileName, Const.BAScriptDataFolder + "\\RemoveBASymLinks.bat");
                    }
                }
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(CopySymLinkBatScripts)}: {e}");
            }
        }
        private void CheckSymLinkAccess()
        {
            try
            {
                string guid = SuperController.singleton.GetTempUID();
                if (FileManagerSecure.DirectoryExists(Const.SymLinksFolder + "\\Custom"))
                {
                    FileManagerSecure.WriteAllBytes(Const.SymLinksFolder + "\\Custom\\" + guid, new byte[1]);
                    if (FileManagerSecure.FileExists("Custom\\" + guid))
                    {
                        Globals.customFolderAccess = true;
                        FileManagerSecure.DeleteFile(Const.SymLinksFolder + "\\Custom\\" + guid);
                    }
                }
                if (FileManagerSecure.DirectoryExists(Const.SymLinksFolder + "\\Saves"))
                {
                    FileManagerSecure.WriteAllBytes(Const.SymLinksFolder + "\\Saves\\" + guid, new byte[1]);
                    if (FileManagerSecure.FileExists("Saves\\" + guid))
                    {
                        Globals.savesFolderAccess = true;
                        FileManagerSecure.DeleteFile(Const.SymLinksFolder + "\\Saves\\" + guid);
                    }
                }
                if (FileManagerSecure.DirectoryExists(Const.SymLinksFolder + "\\AddonPackagesFilePrefs"))
                {
                    FileManagerSecure.WriteAllBytes(Const.SymLinksFolder + "\\AddonPackagesFilePrefs\\" + guid, new byte[1]);
                    if (FileManagerSecure.FileExists("AddonPackagesFilePrefs\\" + guid))
                    {
                        Globals.addonPackageFilePrefsFolderAccess = true;
                        FileManagerSecure.DeleteFile(Const.SymLinksFolder + "\\AddonPackagesFilePrefs\\" + guid);
                    }
                }
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(CheckSymLinkAccess)}: {e}");
            }
            
        }

        public void OnBindingsListRequested(List<object> bindings)
        {
            try
            {
                int pluginSlot = int.Parse(this.name.Substring(7, this.name.IndexOf('_') - 7));

                bindings.Add(new JSONStorableAction("Open UI", () =>
                {
                    browserUI.SetActive();
                }
                ));

                bindings.Add(new JSONStorableAction("Close UI", () =>
                {
                    browserUI.SetInactive();
                }
                ));
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnBindingsListRequested)}: {e}");
            }
        }

        
        private void OnSceneLoaded()
        {
            try
            {
                if (browserUI!=null && browserUI.isVREnabled) browserUI.SetInactive();
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnSceneLoaded)}: {e}");
            }
        }
        private void OnAtomAdded(Atom atom)
        {
         
            try
            {
                if (browserUI != null) browserUI.resourceActionsUI.OnAtomAdded(atom);
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnAtomAdded)}: {e}");
            }
        }

        private void OnAtomNameUpdate(string oldName, string newName)
        {           
            try
            {
                if (browserUI != null) browserUI.resourceActionsUI.AtomNameUpdated(oldName, newName);
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnAtomNameUpdate)}: {e}");
            }
        }

        private void OnAtomRemoved(Atom atom)
        {
            try
            {
                if (browserUI != null) browserUI.resourceActionsUI.OnAtomRemoved(atom);
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnAtomRemoved)}: {e}");
            }
        }

        public void OnEnable()
        {
            try
            {

            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnEnable)}: {e}");
            }
        }

        public void OnDisable()
        {
            try
            {

            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnDisable)}: {e}");
            }
        }

        public void OnDestroy()
        {
            try
            {
                SuperController.singleton.onAtomUIDRenameHandlers -= atomRenameHandle;
                SuperController.singleton.onAtomRemovedHandlers -= atomRemoveHandle;
                SuperController.singleton.onAtomAddedHandlers -= atomAddedHandle;
                SuperController.singleton.onSceneLoadedHandlers -= sceneLoadedHandle;

                //Deregisters with Acidbubbles KeyBindings plugin
                SuperController.singleton.BroadcastMessage("OnActionsProviderDestroyed", this, SendMessageOptions.DontRequireReceiver);

                if (browserUI!=null) browserUI.OnDestroy();
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(OnDestroy)}: {e}");
            }
        }


        void Start()
        {
            try
            {

            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(Start)}: {e}");
            }
        }

        void Update()
        {
            try
            {
                if (browserUI != null) browserUI.Update();
                VARPackageManifest.Update();
                ResourceManifest.Update();
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(BrowserAssist)}.{nameof(Update)}: {e}");
            }

        }
    }

}
