using System.Collections.Generic;
using UnityEngine;

namespace JayJayWon
{
    public static class EnumManifest
    {
        private static Dictionary<string, List<int>> enumManifestDict;
        private static Dictionary<string, string> enumValStoreValDict;
        private static Dictionary<string, string> enumValAltDisplayValDict;
        private static Dictionary<string, int> storeValEnumValDict;
        private static Dictionary<string, int> altDisplayValEnumValDict;

        static EnumManifest()
        {
            enumManifestDict = new Dictionary<string, List<int>>();
            enumValStoreValDict = new Dictionary<string, string>();
            enumValAltDisplayValDict = new Dictionary<string, string>();
            storeValEnumValDict = new Dictionary<string, int>();
            altDisplayValEnumValDict = new Dictionary<string, int>();
        }

        static public void AddEnum(string enumName, int enumVal)
        {
            string enumValKey = EnumValKey(enumName, enumVal);
            if (!enumManifestDict.ContainsKey(enumName)) enumManifestDict.Add(enumName, new List<int>());
            if (!enumManifestDict[enumName].Contains(enumVal)) enumManifestDict[enumName].Add(enumVal);
            else SuperController.LogError("UIAssist.EnumManifest.AddEnum 1: Attempt to add existing enumeration value - " + enumValKey);
        }

        static private string EnumValKey(string enumName, int enumVal)
        {
            return enumName + "." + enumVal.ToString();
        }
        static private string EnumLabelKey(string enumName, string enumLabel)
        {
            return enumName + "." + enumLabel;
        }

        static public void AddEnum(string enumName, int enumVal, string storeValLabel)
        {
            AddEnum(enumName, enumVal);
            string enumValKey = EnumValKey(enumName, enumVal);
            string enumStoreLabelKey = EnumLabelKey(enumName, storeValLabel);
            if (!enumValStoreValDict.ContainsKey(enumName))
            {
                enumValStoreValDict.Add(enumValKey, storeValLabel);
                storeValEnumValDict.Add(enumStoreLabelKey, enumVal);
            }
            else SuperController.LogError("UIAssist.EnumManifest.AddEnum 2: Attempt to add existing enumeration value - " + enumValKey);
        }

        static public void AddEnum(string enumName, int enumVal, string storeValLabel, string altLabel)
        {
            AddEnum(enumName, enumVal, storeValLabel);
            string enumValKey = EnumValKey(enumName, enumVal);
            string enumAltLabelKey = EnumLabelKey(enumName, altLabel);
            if (altLabel != null)
            {
                if (!enumValAltDisplayValDict.ContainsKey(enumName))
                {
                    enumValAltDisplayValDict.Add(enumValKey, altLabel);
                    altDisplayValEnumValDict.Add(enumAltLabelKey, enumVal);
                }
                else SuperController.LogError("UIAssist.EnumManifest.AddEnum 3: Attempt to add existing enumeration value - " + enumValKey);
            }

        }

        static public bool Exists(string enumName, int enumVal)
        {
            if (!enumManifestDict.ContainsKey(enumName)) return false;
            if (!enumManifestDict[enumName].Contains(enumVal)) return false;
            return (true);
        }

        static public string GetDisplayVal(string enumName, int enumVal)
        {
            string enumValKey = EnumValKey(enumName, enumVal);
            if (Exists(enumName, enumVal))
            {
                if (enumValAltDisplayValDict.ContainsKey(enumValKey)) return enumValAltDisplayValDict[enumValKey];
                if (enumValStoreValDict.ContainsKey(enumValKey)) return enumValStoreValDict[enumValKey];
            }
            else SuperController.LogError("UIAssist.EnumManifest.GetLabel: Attempt to retrieve display val for non-existing enumeration value - " + enumValKey);
            return ("");
        }
        static public string GetStoreVal(string enumName, int enumVal)
        {
            string enumValKey = EnumValKey(enumName, enumVal);
            if (Exists(enumName, enumVal))
            {
                if (enumValStoreValDict.ContainsKey(enumValKey)) return enumValStoreValDict[enumValKey];
            }
            else SuperController.LogError("UIAssist.EnumManifest.GetStoreVal: Attempt to retrieve Store Value for non-existing enumeration value - " + enumValKey);
            return ("");
        }


        static public int GetEnumValFromStoreVal(string enumName, string storeValLabel)
        {
            string enumStoreLabelKey = EnumLabelKey(enumName, storeValLabel);
            if (storeValEnumValDict.ContainsKey(enumStoreLabelKey)) return storeValEnumValDict[enumStoreLabelKey];
            else SuperController.LogError("UIAssist.EnumManifest.GetEnumValFromStoreVal: Attempt to retrieve enumeration Value for non-existing enumeration Label - " + enumStoreLabelKey);
            return (-1);
        }

        static public int GetEnumValFromDisplayVal(string enumName, string displayVal)
        {
            string enumStoreLabelKey = EnumLabelKey(enumName, displayVal);
            if (altDisplayValEnumValDict.ContainsKey(enumStoreLabelKey)) return altDisplayValEnumValDict[enumStoreLabelKey];
            else if (storeValEnumValDict.ContainsKey(enumStoreLabelKey)) return storeValEnumValDict[enumStoreLabelKey];
            else SuperController.LogError("UIAssist.EnumManifest.GetEnumValFromDisplayVal: Attempt to retrieve enumeration Value for non-existing enumeration Display Val - " + enumStoreLabelKey);
            return (-1);
        }

        public static List<string> GetDisplayValList(string enumName)
        {
            List<string> displayVals = new List<string>();
            if (enumManifestDict.ContainsKey(enumName))
            {
                foreach (KeyValuePair<string, string> kvp in enumValStoreValDict)
                {
                    if (kvp.Key.StartsWith(enumName + "."))
                    {
                        if (enumValAltDisplayValDict.ContainsKey(kvp.Key)) displayVals.Add(enumValAltDisplayValDict[kvp.Key]);
                        else displayVals.Add(enumValStoreValDict[kvp.Key]);
                    }
                }
            }

            return (displayVals);
        }

        public static List<int> GetEnumsList(string enumName)
        {
            if (enumName == null) return null;
            List<int> enums = new List<int>();
            if (enumManifestDict.ContainsKey(enumName))
            {
                foreach (int enumVal in enumManifestDict[enumName])
                {

                    enums.Add(enumVal);
                }
            }

            return (enums);
        }
        public static SortedDictionary<int, string> GetEnumsDisplayValDict(string enumName)
        {
            if (enumName == null) return null;
            SortedDictionary<int, string> enumsDict = new SortedDictionary<int, string>();
            if (enumManifestDict.ContainsKey(enumName))
            {
                foreach (int enumVal in enumManifestDict[enumName])
                {
                    enumsDict.Add(enumVal, GetDisplayVal(enumName, enumVal));
                }
            }

            return (enumsDict);
        }

        public static Dictionary<int, string> GetEnumsStoreValDict(string enumName)
        {
            if (enumName == null) return null;
            Dictionary<int, string> enumsDict = new Dictionary<int, string>();
            if (enumManifestDict.ContainsKey(enumName))
            {
                foreach (int enumVal in enumManifestDict[enumName])
                {
                    enumsDict.Add(enumVal, GetStoreVal(enumName, enumVal));
                }
            }

            return (enumsDict);
        }

        public static Dictionary<string, int> GetStoreValEnumsDict(string enumName)
        {
            if (enumName == null) return null;
            Dictionary<string, int> enumsDict = new Dictionary<string, int>();
            if (enumManifestDict.ContainsKey(enumName))
            {
                foreach (int enumVal in enumManifestDict[enumName])
                {
                    enumsDict.Add(GetStoreVal(enumName, enumVal), enumVal);
                }
            }

            return (enumsDict);
        }
        public static Dictionary<string, int> GetDisplayValEnumsDict(string enumName)
        {
            if (enumName == null) return null;
            Dictionary<string, int> enumsDict = new Dictionary<string, int>();
            if (enumManifestDict.ContainsKey(enumName))
            {
                foreach (int enumVal in enumManifestDict[enumName])
                {
                    enumsDict.Add(GetDisplayVal(enumName, enumVal), enumVal);
                }
            }

            return (enumsDict);
        }

    }

    public static class DesktopDisplayMode
    {
        public const int rightDocked = 1;
        public const int fullWidth = 2;
        public const int leftDocked = 3;
        public const int leftMinimized = 4;
        public const int rightMinimized = 5;
    }

    public static class ActionUITabType
    {
        public const int primaryActionTab = 1;
        public const int tagsTab = 2;

        private static Dictionary<int, string> actionUITabTypeLabels = new Dictionary<int, string>();

        public static string GetTabLabel(int tabType)
        {
            return actionUITabTypeLabels[tabType];
        }
        static ActionUITabType()
        {
            actionUITabTypeLabels.Add(primaryActionTab, "Action");
            actionUITabTypeLabels.Add(tagsTab, "Tags");

        }
    }

    public static class LicenseType
    {
        public const int Unknown = -1;
        public const int Any = 0;
        public const int PC = 1;
        public const int PCEA = 2;
        public const int Questionable = 3;
        public const int FC = 4;
        public const int CCBY = 5;
        public const int CCBYSA = 6;
        public const int CCBYND = 7;
        public const int CCBYNC = 8;
        public const int CCBYNCSA = 9;
        public const int CCBYNCND = 10;

        public static string enumManifestName
        {
            get { return "LicenseType"; }
        }
        static LicenseType()
        {
            EnumManifest.AddEnum(enumManifestName, Unknown, "Unknown", "Unknown");
            EnumManifest.AddEnum(enumManifestName, Any, "Any", "<i><b>Any</b></i>");
            EnumManifest.AddEnum(enumManifestName, PC, "PC", "PC");
            EnumManifest.AddEnum(enumManifestName, PCEA, "PC EA", "PC EA");
            EnumManifest.AddEnum(enumManifestName, Questionable, "Questionable", "Questionable");
            EnumManifest.AddEnum(enumManifestName, FC, "FC", "FC");
            EnumManifest.AddEnum(enumManifestName, CCBY, "CC BY", "CC BY");

            EnumManifest.AddEnum(enumManifestName, CCBYSA, "CC BY-SA", "CC BY-SA");
            EnumManifest.AddEnum(enumManifestName, CCBYND, "CC BY-ND", "CC BY-ND");
            EnumManifest.AddEnum(enumManifestName, CCBYNC, "CC BY-NC", "CC BY-NC");
            EnumManifest.AddEnum(enumManifestName, CCBYNCSA, "CC BY-NC-SA", "CC BY-NC-SA");
            EnumManifest.AddEnum(enumManifestName, CCBYNCND, "CC BY-NC-ND", "CC BY-NC-ND");
        }
    }
    public static class LeftRight
    {
        public const int left = 1;
        public const int right = 2;
        public const int both = 3;
        public const int neither = 4;

        public static string enumManifestName
        {
            get { return "LeftRight"; }
        }
        static LeftRight()
        {
            EnumManifest.AddEnum(enumManifestName, left);
            EnumManifest.AddEnum(enumManifestName, right);
            EnumManifest.AddEnum(enumManifestName, both);
            EnumManifest.AddEnum(enumManifestName, neither);
        }
    }

    public static class ResourceTargetCategory
    {
        public const int nonAtomTargetable = 1;
        public const int personAtomTargetable = 2;
        public const int otherAtomTargetable = 3;
        public const int anyAtomTargetable = 4;
        public const int textureAtomTargetable = 5;
        public const int anyTarget = 6;
    }
    public static class ResourceCategory
    {
        public const int preset = 1;
        public const int scene = 2;
        public const int plugins = 2;
    }

    public static class ResourceType
    {
        public const int unknown = 0;
        public const int scene = 1;
        public const int presetAppearance = 2;
        public const int presetAnimation = 3;
        public const int presetBreastPhys = 4;
        public const int presetClothing = 5;
        public const int presetGeneral = 6;
        public const int presetGlutePhysics = 7;
        public const int presetHair = 8;
        public const int presetMorphs = 9;
        public const int presetPlugins = 10;
        public const int presetPose = 11;
        public const int presetSkin = 12;
        public const int personTexture = 13;
        public const int atomTexture = 14;
        public const int plugins = 15;

        private static Dictionary<int, int> resourceTypeToTargetCategory = new Dictionary<int, int>();
        private static Dictionary<int, int> resourceTypeToCategory = new Dictionary<int, int>();
        private static Dictionary<int, HashSet<int>> resourceCategoryToType = new Dictionary<int, HashSet<int>>();
        private static Dictionary<int, HashSet<int>> resourceTargetCategoryToType = new Dictionary<int, HashSet<int>>();
        private static Dictionary<int, HashSet<int>> resourceTypeToTagCategory = new Dictionary<int, HashSet<int>>();

        public static string enumManifestName
        {
            get { return "ResourceType"; }
        }
        static ResourceType()
        {
            EnumManifest.AddEnum(enumManifestName, scene,"Scene");
            AssignCategories(scene, ResourceCategory.scene, ResourceTargetCategory.nonAtomTargetable);
            AssignTags(scene, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetAppearance, "Preset Appearance");
            AssignCategories(presetAppearance, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetAppearance, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetAnimation, "Preset Animation");
            AssignCategories(presetAnimation, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetAnimation, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetBreastPhys, "Preset Breast Physics");
            AssignCategories(presetBreastPhys, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetBreastPhys, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetClothing, "Preset Clothing");
            AssignCategories(presetClothing, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetClothing, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetGlutePhysics, "Preset Glute Physics");
            AssignCategories(presetGlutePhysics, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetGlutePhysics, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetGeneral, "Preset General");
            AssignCategories(presetGeneral, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetGeneral, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetHair, "Preset Hair");
            AssignCategories(presetHair, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetHair, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetMorphs, "Preset Morphs");
            AssignCategories(presetMorphs, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetMorphs, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetPlugins, "Preset Plugins");
            AssignCategories(presetPlugins, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetPlugins, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetPose, "Preset Pose");
            AssignCategories(presetPose, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetPose, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, presetSkin, "Preset Skin");
            AssignCategories(presetSkin, ResourceCategory.preset, ResourceTargetCategory.personAtomTargetable);
            AssignTags(presetSkin, TagCategory.gender, TagCategory.userDefined);

            EnumManifest.AddEnum(enumManifestName, plugins, "Plugins");
            AssignCategories(plugins, ResourceCategory.plugins, ResourceTargetCategory.anyTarget);
            AssignTags(plugins, TagCategory.userDefined);

            //            EnumManifest.AddEnum(enumManifestName, personTexture, "Person Texture");

            //            EnumManifest.AddEnum(enumManifestName, atomTexture, "Atom Texture");


        }

        private static void AssignTags (int resourceType, params int[] tagCategories)
        {
            if (!resourceTypeToTagCategory.ContainsKey(resourceType)) resourceTypeToTagCategory.Add(resourceType, new HashSet<int>());

            foreach (int tagCat in tagCategories)
            {
                resourceTypeToTagCategory[resourceType].Add(tagCat);
            }
        }
        private static void AssignCategories(int resourceType, int resourceCategory, int targetCategory)
        {
            resourceTypeToCategory.Add(resourceType, resourceCategory);
            resourceTypeToTargetCategory.Add(resourceType, targetCategory);

            if (!resourceCategoryToType.ContainsKey(resourceCategory)) resourceCategoryToType.Add(resourceCategory, new HashSet<int>());
            resourceCategoryToType[resourceCategory].Add(resourceType);

            if (!resourceTargetCategoryToType.ContainsKey(targetCategory)) resourceTargetCategoryToType.Add(targetCategory, new HashSet<int>());
            resourceTargetCategoryToType[targetCategory].Add(resourceType);
        }

        public static int GetResourceTargetCategory(int resourceType)
        {
            return resourceTypeToTargetCategory[resourceType];
        }
        public static int GetResourceCategory(int resourceType)
        {
            if (resourceTypeToCategory.ContainsKey(resourceType)) return resourceTypeToCategory[resourceType];
            return -1;
        }
        public static HashSet<int> GetTagCategories(int resourceType)
        {
            if (!resourceTypeToTagCategory.ContainsKey(resourceType)) resourceTypeToTagCategory.Add(resourceType, new HashSet<int>());
            return resourceTypeToTagCategory[resourceType];
        }

        public static HashSet<int> GetResourceTypesByCategory(int resourceCategory)
        {
            if (!resourceCategoryToType.ContainsKey(resourceCategory)) resourceCategoryToType.Add(resourceCategory, new HashSet<int>());
            return resourceCategoryToType[resourceCategory];
        }

        public static HashSet<int> GetResourceTypesByTargetCategory(int resourceCategory)
        {
            if (!resourceTargetCategoryToType.ContainsKey(resourceCategory)) resourceTargetCategoryToType.Add(resourceCategory, new HashSet<int>());
            return resourceTargetCategoryToType[resourceCategory];
        }
    }

    public static class CreatorFilterChoices
    {
        public const int all = 3;
        public const int allVARs = 1;
        public const int localOnly = 2;

        public static string enumManifestName
        {
            get { return "CreatorFilterChoices"; }
        }
        static CreatorFilterChoices()
        {
            EnumManifest.AddEnum(enumManifestName, allVARs,"Package Creators", "<i><b>Package Creators</b></i>");
            EnumManifest.AddEnum(enumManifestName, localOnly, "Local Folders", "<i><b>Local Folders</b></i>");
            EnumManifest.AddEnum(enumManifestName, all,"Local & Packaged Content", "<i><b>Local & Packaged</b></i>");

        }
    }

    public static class BrowserSelectMode
    {
        public const int instant = 1;
        public const int single = 2;
        public const int multi = 3;

        public static string enumManifestName
        {
            get { return "BrowserSelectMode"; }
        }
        static BrowserSelectMode()
        {
            EnumManifest.AddEnum(enumManifestName, instant, "Instant");
            EnumManifest.AddEnum(enumManifestName, single, "Single");
            EnumManifest.AddEnum(enumManifestName, multi, "Multi");
        }
    }

    public static class BrowserSortMode
    {
        public const int AtoZ = 1;
        public const int ZtoA = 2;
        public const int packageFolder = 3;
        public const int cataloguedDateTime = 4;

        public static string enumManifestName
        {
            get { return "BrowserSortMode"; }
        }
        static BrowserSortMode()
        {
            EnumManifest.AddEnum(enumManifestName, AtoZ, "A to Z");
            EnumManifest.AddEnum(enumManifestName, ZtoA, "Z to A");
            EnumManifest.AddEnum(enumManifestName, packageFolder, "Package & Folder");
            EnumManifest.AddEnum(enumManifestName, cataloguedDateTime, "Catalogued Date");

        }
    }
    public static class PluginTargetTypes
    {
        public const int sessionPlugins = 1;
        public const int scenePlugins = 2;

        public static string enumManifestName
        {
            get { return "PluginTargetTypes"; }
        }
        static PluginTargetTypes()
        {
            EnumManifest.AddEnum(enumManifestName, sessionPlugins, "Session Plugins");
            EnumManifest.AddEnum(enumManifestName, scenePlugins, "Scene Plugins");
        }
    }

    public static class NoneAvailable
    {
        public const int noneAvailable = 1;
        public static string enumManifestName
        {
            get { return "NoneAvailable"; }
        }
        static NoneAvailable()
        {
            EnumManifest.AddEnum(enumManifestName, noneAvailable, "<None Available>");
        }
    }

    public static class TriState
    {
        public const int active = 1;
        public const int inactive = 2;
        public const int indeterminate = 3;
        public static string enumManifestName
        {
            get { return "TriState"; }
        }
        static TriState()
        {
            EnumManifest.AddEnum(enumManifestName, active, "Active");
            EnumManifest.AddEnum(enumManifestName, inactive, "Inactive");
            EnumManifest.AddEnum(enumManifestName, indeterminate, "Indeterminate");
        }
    }

    public static class TagCategory
    {
        public const int all = 0;
        public const int bodyRegion = 1;
        public const int type = 2;
        public const int gender = 3;
        public const int userDefined = 4;

        private static Dictionary<int, Color> tagCategoryColor = new Dictionary<int, Color>();
        public static string enumManifestName
        {
            get { return "TagCategory"; }
        }
        static TagCategory()
        {
            EnumManifest.AddEnum(enumManifestName, all, "All");
            EnumManifest.AddEnum(enumManifestName, bodyRegion, "Body Region");
            tagCategoryColor.Add(bodyRegion, new Color(0.75f, 0.75f, 1f));
            EnumManifest.AddEnum(enumManifestName, type, "Type");
            tagCategoryColor.Add(type, new Color(0.75f, 1f, 0.75f));
            EnumManifest.AddEnum(enumManifestName, gender, "Gender");
            tagCategoryColor.Add(gender, new Color(1f,0.75f,0.75f));
            EnumManifest.AddEnum(enumManifestName, userDefined, "User");
            tagCategoryColor.Add(userDefined, new Color(0.75f, 1f, 1f));
        }

        public static Color GetColor(int tagCategory)
        {
            return tagCategoryColor[tagCategory];
        }
    }

    public static class TagGroups
    {
        public const int anyTags = 1;
        public const int hiddenOnly = 2;
        public const int noTags = 3;
        public const int noUserTags = 4;
        public static string enumManifestName
        {
            get { return "TagGroups"; }
        }
        static TagGroups()
        {
            EnumManifest.AddEnum(enumManifestName, anyTags, "Any Tags", "<i><b>Any Tags</b></i>");
            EnumManifest.AddEnum(enumManifestName, hiddenOnly, "Only Hidden", "<i><b>Only Hidden</b></i>");
            EnumManifest.AddEnum(enumManifestName, noTags, "No Tags", "<i><b>No Tags</b></i>");
            EnumManifest.AddEnum(enumManifestName, noUserTags, "No User Tags", "<i><b>No User Tags</b></i>");
        }
    }

}
