﻿using System.Collections.Generic;
using System.Linq;


namespace JayJayWon
{
    public class Tag
    {
        protected string _tagName;
        public JSONStorableEnumStringChooser tagCategoryJSEnum;
        protected int _resourceType;


        public string tagName { get
            {
                return _tagName;
            }
        }

        public int tagCategory
        {
            get
            {
                return tagCategoryJSEnum.val;
            }
        }

        public int resourceType
        {
            get
            {
                return _resourceType;
            }
        }


        public Tag(string newTagName, int tagCategory, int newTagResourceType)
        {
            tagCategoryJSEnum = new JSONStorableEnumStringChooser("tagCategory", TagCategory.enumManifestName, tagCategory, "Tag Category");

            _tagName = newTagName;
            _resourceType = newTagResourceType;
        }
    }


    public static class TagManifest
    {
        private static HashSet<Tag> tags = new HashSet<Tag>();

        private static Dictionary<ResourceVersionGroupEntry, HashSet<Tag>> rvgeTags = new Dictionary<ResourceVersionGroupEntry, HashSet<Tag>>();
        private static Dictionary<Tag, HashSet<ResourceVersionGroupEntry>> tagsRVGEs = new Dictionary<Tag, HashSet<ResourceVersionGroupEntry>>();

        private static Dictionary<int, Dictionary<string, Tag>> tagNamesToTagByResourceType = new Dictionary<int, Dictionary<string, Tag>>();

        private static Dictionary<int, HashSet<Tag>> categoryTagMap = new Dictionary<int, HashSet<Tag>>();

        private static Dictionary<int, HashSet<Tag>> resourceTypeTagMap = new Dictionary<int, HashSet<Tag>>();

        private static Dictionary<int, HashSet<int>> resourceTypeTagCategoryMap = new Dictionary<int, HashSet<int>>();

        static TagManifest()
        {
            foreach (int resourceType in ResourceType.GetResourceTypesByTargetCategory(ResourceTargetCategory.personAtomTargetable))
            {
                CreateTag("male", TagCategory.gender, resourceType);
                CreateTag("female", TagCategory.gender, resourceType);
            }
            
        }


        public static Tag CreateTag(string name, int tagCategory, int resourceType)
        {
            if (tags.Any(t => t.tagName == name && t.tagCategory==tagCategory && t.resourceType == resourceType)) return null;

            Tag newTag = new Tag(name, tagCategory, resourceType);
            tags.Add(newTag);
            if (!categoryTagMap.ContainsKey(tagCategory)) categoryTagMap.Add(tagCategory, new HashSet<Tag>());
            categoryTagMap[tagCategory].Add(newTag);

            if (!resourceTypeTagMap.ContainsKey(resourceType)) resourceTypeTagMap.Add(resourceType, new HashSet<Tag>());
            resourceTypeTagMap[resourceType].Add(newTag);

            if (!resourceTypeTagCategoryMap.ContainsKey(resourceType)) resourceTypeTagCategoryMap.Add(resourceType, new HashSet<int>());
            resourceTypeTagCategoryMap[resourceType].Add(newTag.tagCategory);

            if (!tagNamesToTagByResourceType.ContainsKey(resourceType)) tagNamesToTagByResourceType.Add(resourceType, new Dictionary<string, Tag>());

            var tagNamesToTag = tagNamesToTagByResourceType[resourceType];
            if (!tagNamesToTag.ContainsKey(name)) tagNamesToTag.Add(name, newTag);
            else tagNamesToTag[name] = newTag;

            return newTag;
        }

        public static HashSet<Tag> GetRVGETags(ResourceVersionGroupEntry rvge)
        {
            if (!rvgeTags.ContainsKey(rvge)) return new HashSet<Tag>();
            return rvgeTags[rvge];
        }

        public static List<Tag> GetTags(int resourceType)
        {
            if (!resourceTypeTagMap.ContainsKey(resourceType)) resourceTypeTagMap.Add(resourceType, new HashSet<Tag>());
            return resourceTypeTagMap[resourceType].OrderBy(t=>t.tagCategory).ThenBy(t=>t.tagName).ToList();
        }

        public static List<string> GetTagNames(int resourceType)
        {
            if (!tagNamesToTagByResourceType.ContainsKey(resourceType)) tagNamesToTagByResourceType.Add(resourceType, new Dictionary<string, Tag>());
            return tagNamesToTagByResourceType[resourceType].Keys.ToList();
        }

        public static HashSet<int> GetTagCategories(int resourceType)
        {
            if (!resourceTypeTagCategoryMap.ContainsKey(resourceType)) resourceTypeTagCategoryMap.Add(resourceType, new HashSet<int>());
            return resourceTypeTagCategoryMap[resourceType];
        }

        public static Tag GetTag(int resourceType, string tagName)
        {
            return tagNamesToTagByResourceType[resourceType][tagName];
        }

        public static bool IsRVGETagged(ResourceVersionGroupEntry rvge, Tag tag)
        {
            if (!rvgeTags.ContainsKey(rvge)) return false;

            if (rvgeTags[rvge].Contains(tag)) return true;

            return false;
        }

        public static bool IsRVGENotTagged(ResourceVersionGroupEntry rvge)
        {
            if (!rvgeTags.ContainsKey(rvge)) return true;

            if (rvgeTags[rvge].Count==0) return true;

            return false;
        }

        public static bool IsRVGENotUserTagged(ResourceVersionGroupEntry rvge)
        {
            if (!rvgeTags.ContainsKey(rvge)) return true;

            if (rvgeTags[rvge].Where(t=>t.tagCategory==TagCategory.userDefined).Count() == 0) return true;

            return false;
        }


        public static void SetRVGETagged(ResourceVersionGroupEntry rvge, Tag tag, bool state)
        {           

            if (state)
            {
                if (!rvgeTags.ContainsKey(rvge)) rvgeTags.Add(rvge, new HashSet<Tag>());
                rvgeTags[rvge].Add(tag);
                if (!tagsRVGEs.ContainsKey(tag)) tagsRVGEs.Add(tag, new HashSet<ResourceVersionGroupEntry>());
                tagsRVGEs[tag].Add(rvge);

            }
            else
            {               
                if (!rvgeTags.ContainsKey(rvge)) return;
                rvgeTags[rvge].Remove(tag);
                tagsRVGEs[tag].Remove(rvge);
            }
        }

        public static void SetRVGETagged(ResourceVersionGroupEntry rvge, string tagName, bool state)
        {
            Tag tag = tagNamesToTagByResourceType[rvge.resourceType][tagName];
            SetRVGETagged(rvge, tag, state);
        }

    }
}
