using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.UI;
using UnityEngine.EventSystems;

namespace JayJayWon
{

    public static class UI
    {
        public static GameObject CreateVScrollableContent(GameObject parentGO, Color backgroundColor, int anchorPreset, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset)
        {
            GameObject scrollableContentGO = AddChildGOImage(parentGO, backgroundColor, anchorPreset, horizontalSize, verticalSize, anchoredPositionOffset);

            GameObject buttonListViewportGO = CreateVScrollableContentViewPort(scrollableContentGO, horizontalSize - 10f, verticalSize);
            buttonListViewportGO.name = "scrollableContentViewportGO";

            GameObject buttonListScrollbarGO = CreateScrollBar(scrollableContentGO, 10f, verticalSize, 0f, 0f, Scrollbar.Direction.BottomToTop);
            buttonListScrollbarGO.name = "scollBar";
            

            ScrollRect scrollRect = scrollableContentGO.AddComponent<ScrollRect>();
            GameObject buttonListContentGO = buttonListViewportGO.transform.Find("vScrollableContentGO").gameObject;
            scrollRect.content = buttonListContentGO.GetComponent<RectTransform>();
            scrollRect.horizontal = false;
            scrollRect.vertical = true;

            scrollRect.movementType = ScrollRect.MovementType.Clamped;
            scrollRect.viewport = buttonListViewportGO.GetComponent<RectTransform>();
            scrollRect.verticalScrollbar = buttonListScrollbarGO.GetComponent<Scrollbar>();
            scrollRect.verticalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;

            return (scrollableContentGO);
        }

        private static GameObject CreateScrollBar(GameObject parentGO, float width, float height, float xAnchorOffset, float yAnchorOffset, Scrollbar.Direction scrollbarDirection)
        {
            Sprite backgroundSprite = Resources.FindObjectsOfTypeAll<Sprite>().First(sprite => sprite.name == "Background");
            Sprite uiSprite = Resources.FindObjectsOfTypeAll<Sprite>().First(sprite => sprite.name == "UISprite");

            GameObject scrollbarGO = AddChildGOImage(parentGO, new Color(1f, 1f, 1f), AnchorPresets.vStretchRight, width, height, Vector2.zero, backgroundSprite);
            Scrollbar scrollbar = scrollbarGO.AddComponent<Scrollbar>();
            scrollbar.SetDirection(scrollbarDirection, false);
            scrollbarGO.SetActive(true);

            GameObject scrollbarSlidingAreaGO = new GameObject();
            scrollbarSlidingAreaGO.AddComponent<RectTransform>();
            scrollbarSlidingAreaGO.transform.SetParent(scrollbarGO.transform, false);
            scrollbarSlidingAreaGO.SetActive(true);
            ResizeRecTransform(scrollbarSlidingAreaGO, width, height, 0f, 0f, AnchorPresets.vStretchRight);

            GameObject scrollbarHandleGO = AddChildGOImage(scrollbarSlidingAreaGO, new Color(1f, 1f, 1f), AnchorPresets.centre, 3f, 10f, Vector2.zero, uiSprite);

            scrollbar.targetGraphic = scrollbarHandleGO.GetComponent<UnityEngine.UI.Image>();
            scrollbar.handleRect = scrollbarHandleGO.GetComponent<RectTransform>();

            return (scrollbarGO);
        }


        private static GameObject CreateVScrollableContentViewPort(GameObject parentGO, float width, float height)
        {
            GameObject scrollableContentViewportGO = AddChildGOImage(parentGO, new Color(1f, 0f, 0f,0f), AnchorPresets.vStretchLeft, width, height, Vector2.zero);
 //           Mask blvpMask = scrollableContentViewportGO.AddComponent<Mask>();
 //           blvpMask.showMaskGraphic = false;
            RectMask2D altMask = scrollableContentViewportGO.AddComponent<RectMask2D>();

            GameObject scrollableContentGO = AddChildGOImage(scrollableContentViewportGO, new Color(0f, 0f, 1f,0f), AnchorPresets.vStretchLeft, 0f, 200f, Vector2.zero, null, false);
            scrollableContentGO.name = "vScrollableContentGO";

            VerticalLayoutGroup vlg = scrollableContentGO.AddComponent<VerticalLayoutGroup>();
            vlg.childAlignment = TextAnchor.UpperLeft;
            vlg.childForceExpandHeight = true;
            vlg.childForceExpandWidth = true;
            vlg.childControlHeight = false;
            vlg.childControlWidth = false;
            vlg.spacing = 0f;

            ContentSizeFitter csf = scrollableContentGO.AddComponent<ContentSizeFitter>();
            csf.horizontalFit = ContentSizeFitter.FitMode.MinSize;
            csf.verticalFit = ContentSizeFitter.FitMode.MinSize;

            return (scrollableContentViewportGO);
        }

        public static void AddImageComponent(GameObject go, Color color, Sprite sprite = null, bool enabled = true)
        {
            UnityEngine.UI.Image image = go.AddComponent<UnityEngine.UI.Image>();
            image.color = color;
            if (sprite!=null) image.sprite = sprite;
            image.type = UnityEngine.UI.Image.Type.Sliced;
            image.enabled = enabled;
            
        }

        public static GameObject AddChildImage(Transform parentTransform, Color color, Vector2 anchorMin, Vector2 anchorMax, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset, Vector2 pivot, Sprite sprite = null, bool enabled = true)
        {
            GameObject go = new GameObject();
            go.transform.SetParent(parentTransform, false);

            AddImageComponent(go, color, sprite, enabled);
            RectTransform rt = go.GetComponent<RectTransform>();
            rt.sizeDelta = Vector2.zero;
            rt.anchorMin = anchorMin;
            rt.anchorMax = anchorMax;
            rt.pivot = pivot;
            if (verticalSize != 0f) rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, verticalSize);
            if (horizontalSize != 0f) rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, horizontalSize);

            rt.anchoredPosition = anchoredPositionOffset;

            return go;
        }

        public static GameObject AddChildImage(GameObject parentGO, Color color, Vector2 anchorMin, Vector2 anchorMax, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset, Vector2 pivot, Sprite sprite = null, bool enabled = true)
        {
            return AddChildImage(parentGO.transform, color, anchorMin, anchorMax, horizontalSize, verticalSize, anchoredPositionOffset, pivot, sprite, enabled);
        }
        public static GameObject AddChildGOImage(GameObject parentGO, Color color, int anchorPreset, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset, Sprite sprite = null, bool enabled = true)
        {
            return AddChildImage(parentGO.transform, color, AnchorPresets.GetAnchorMin(anchorPreset), AnchorPresets.GetAnchorMax(anchorPreset), horizontalSize, verticalSize, anchoredPositionOffset, AnchorPresets.GetPivot(anchorPreset), sprite, enabled);
        }

        public static GameObject AddChildImage(Transform parentTransform, Color color, int anchorPreset, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset, Sprite sprite = null, bool enabled = true)
        {
            return AddChildImage(parentTransform, color, AnchorPresets.GetAnchorMin(anchorPreset), AnchorPresets.GetAnchorMax(anchorPreset), horizontalSize, verticalSize, anchoredPositionOffset, AnchorPresets.GetPivot(anchorPreset), sprite, enabled);
        }

        public static GameObject CreateBasicTextBox(GameObject parentGO, JSONStorableString jss, float width, float height, int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool resizeTextForBestFit = false, int minFontSize = 20, HorizontalWrapMode horizontalWrapMode = HorizontalWrapMode.Wrap, Color? backgoundColor=null)
        {
            GameObject imageGO = AddChildGOImage(parentGO, backgoundColor ?? Color.gray, anchorPreset, width, height, new Vector2(anchorXOffset, anchorYOffset));
            var textGo = new GameObject("text");
            textGo.transform.SetParent(imageGO.transform, false);
            textGo.AddComponent<RectTransform>();
            Text text = textGo.AddComponent<Text>();
            Font selectedFont = (Font)Resources.GetBuiltinResource(typeof(Font), "Arial.ttf");
            text.font = selectedFont;
            ReformatText(textGo, width - 4f, height, fontSize, 2f, 0, AnchorPresets.centre, TextAnchor.MiddleLeft, resizeTextForBestFit);

            text.horizontalOverflow = horizontalWrapMode;
            //           text.resizeTextForBestFit = resizeTextForBestFit;

            if (resizeTextForBestFit)
            {
                text.resizeTextMaxSize = fontSize;
                text.resizeTextMinSize = minFontSize;
            }
            else text.fontSize = fontSize;

            jss.RegisterText(text);

            text.color = Color.white;
            return imageGO;
        }

        public static UIDynamicTextField CreateTextBox(GameObject parentGO, JSONStorableString jss, float width, float height, int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool input = false, InputField.LineType inputFieldType = InputField.LineType.MultiLineNewline, string placeHolderText = "", float hScrollBarHeight = 0f, float vScrollBarWidth = 10f, Action<string> textEntryChangeCallback = null)
        {
            Transform transform = GameObject.Instantiate<Transform>(Globals.mvrScript.manager.configurableTextFieldPrefab);

            GameObject textFieldGO = transform.gameObject;
            UIDynamicTextField textFieldUID = textFieldGO.transform.GetComponent<UIDynamicTextField>();
            transform.SetParent(parentGO.transform, false);

            ResizeRecTransform(textFieldGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);
            if (input)
            {
                textFieldUID.backgroundColor = Color.white;
                var inputComponent = textFieldUID.gameObject.AddComponent<InputField>();

                inputComponent.textComponent = textFieldUID.UItext;
                inputComponent.lineType = inputFieldType;

                inputComponent.textComponent.resizeTextForBestFit = false;
                jss.RegisterInputField(inputComponent);

                if (placeHolderText != "")
                {
                    var placeholderGo = new GameObject("placeHolderText");
                    placeholderGo.transform.SetParent(textFieldUID.gameObject.transform, false);

                    var placeholderRect = placeholderGo.AddComponent<RectTransform>();
                    Font selectedFont = (Font)Resources.GetBuiltinResource(typeof(Font), "Arial.ttf");

                    Text placeholderUIText = placeholderGo.AddComponent<Text>();
                    placeholderUIText.raycastTarget = false;
                    placeholderUIText.font = selectedFont;
                    placeholderUIText.color = Color.gray;
                    placeholderUIText.fontStyle = FontStyle.Italic;

                    ReformatText(placeholderGo, width - vScrollBarWidth, height - hScrollBarHeight, fontSize, 2f, 0f, AnchorPresets.middleLeft, TextAnchor.MiddleLeft);

                    if (jss.val == "") placeholderUIText.text = placeHolderText;

                    inputComponent.onValueChanged.AddListener((text) =>
                    {
                        InputTextFieldValueChanged(inputComponent, placeholderUIText, jss, placeHolderText, textEntryChangeCallback);
                    });
                }
            }
            else jss.RegisterText(textFieldUID.UItext);

            textFieldUID.height = height - hScrollBarHeight;

            GameObject scrollViewGO = textFieldGO.transform.Find("Scroll View").gameObject;
            ReformatScrollView(scrollViewGO, width, vScrollBarWidth, height, hScrollBarHeight);

            GameObject contentGO = textFieldGO.transform.Find("Scroll View/Viewport/Content").gameObject;
            ResizeRecTransform(contentGO, width, height, anchorPreset: AnchorPresets.topLeft);
            AddChildGOImage(contentGO, new Color(0, 0f, 0f, 0f), AnchorPresets.topLeft, width, height, new Vector2(0f, 0f));

            GameObject textContentGO = textFieldGO.transform.Find("Scroll View/Viewport/Content/Text").gameObject;
            ReformatText(textContentGO, width - vScrollBarWidth, height - hScrollBarHeight, fontSize, 2f, 0f, anchorPreset: AnchorPresets.centre, textAlginment: TextAnchor.MiddleLeft);

            return textFieldUID;
        }
        private static void InputTextFieldValueChanged(InputField inputField, Text placeholderUIText, JSONStorableString jss, string placeHolderText, Action<string> textEntryChangeCallback)
        {
            if (inputField.text == "") placeholderUIText.text = placeHolderText;
            else placeholderUIText.text = "";

            if (textEntryChangeCallback != null) textEntryChangeCallback(inputField.text);
        }

        public static UIDynamicButton CreateUIButton(GameObject parentGO, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none)
        {
            Transform buttonTransform = UnityEngine.Object.Instantiate(Globals.mvrScript.manager.configurableButtonPrefab.transform, parentGO.transform, false);

            GameObject buttonGO = buttonTransform.gameObject;
            buttonTransform.SetParent(parentGO.transform, false);

            UIDynamicButton buttonUID = buttonGO.transform.GetComponent<UIDynamicButton>();
            Transform textTransform = buttonUID.button.transform.Find("Text");
            ReformatText(textTransform.GetComponent<RectTransform>().gameObject, width, height, fontSize, 0, 0, AnchorPresets.centre);

            buttonUID.label = label;
            buttonUID.buttonText.fontSize = fontSize;

            ResizeRecTransform(buttonGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            return (buttonUID);

        }

        public static UIDynamicToggle CreateToggle(GameObject parentGO, JSONStorableBool jsb, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none)
        {
            Transform transform = GameObject.Instantiate<Transform>(Globals.mvrScript.manager.configurableTogglePrefab);

            GameObject toggleGO = transform.gameObject;
            UIDynamicToggle toggleUID = toggleGO.transform.GetComponent<UIDynamicToggle>();
            toggleUID.label = label;
            jsb.RegisterToggle(toggleUID.toggle);

            transform.SetParent(parentGO.transform, false);

            ResizeRecTransform(toggleGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject labelTextGO = toggleGO.transform.Find("Label").gameObject;
            ReformatText(labelTextGO, width - 20f, height, fontSize, 0f, 0f, AnchorPresets.middleRight, textAlginment: TextAnchor.MiddleLeft);

            GameObject checkBoxGO = toggleGO.transform.Find("Background").gameObject;
            ResizeRecTransform(checkBoxGO, 20f, 20f, 0f, 0f, AnchorPresets.middleLeft);

            GameObject checkMarkGO = toggleGO.transform.Find("Background/Checkmark").gameObject;
            ResizeRecTransform(checkMarkGO, 20f, 20f, 0f, 0f, AnchorPresets.centre);

            return toggleUID;
        }


        public static UIDynamicPopup CreatePopup(GameObject parentGO, JSONStorableStringChooser jssc, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool popupAbove=false)
        {
            float labelProportion = 0.4f;
            Transform transform = GameObject.Instantiate<Transform>(Globals.mvrScript.manager.configurablePopupPrefab);

            GameObject popupGO = transform.gameObject;
            UIDynamicPopup uiPopup = popupGO.transform.GetComponent<UIDynamicPopup>();
            uiPopup.label = label;
            uiPopup.popup.selectColor = Color.cyan;

            jssc.RegisterPopup(uiPopup.popup);

            transform.SetParent(parentGO.transform, false);

            ResizeRecTransform(popupGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject dropDownButtonGO = transform.Find("Button").gameObject;
            ReformatUnityButton(dropDownButtonGO, width * (1 - labelProportion), height * .9f, fontSize, 0f, 0f, AnchorPresets.middleRight);

            GameObject textGO = transform.Find("Text").gameObject;
            ReformatText(textGO, width * (labelProportion), height * .9f, fontSize, 0f, 0f, AnchorPresets.middleLeft);

            GameObject popupPanelGO = transform.Find("PopupPanel").gameObject;
            popupPanelGO.transform.Translate(new Vector3(0, 0, 0.06f));

            if (popupAbove) ReformatPopupPanel(popupPanelGO, width * (1 - labelProportion), 300f, fontSize, 0f, height, anchorPreset: AnchorPresets.bottomRight);
            else ReformatPopupPanel(popupPanelGO, width * (1 - labelProportion), 300f, fontSize, 0f, -height, anchorPreset: AnchorPresets.topRight);

            return uiPopup;
        }
        public static UIDynamicPopup CreateScrollablePopup(GameObject parentGO, JSONStorableStringChooser jssc, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, float popupHeight = 300f, bool popupAbove = false, bool disableSlider = false, int anchorPreset = AnchorPresets.none)
        {
            float labelProportion = 0.4f;
            Transform transform = GameObject.Instantiate<Transform>(Globals.mvrScript.manager.configurableScrollablePopupPrefab);

            GameObject popupGO = transform.gameObject;
            UIDynamicPopup uiPopup = popupGO.transform.GetComponent<UIDynamicPopup>();
            uiPopup.label = label;
            
            jssc.RegisterPopup(uiPopup.popup);

            float sliderHeight = 10f;
            if (disableSlider) sliderHeight = 0f;

            uiPopup.popup.selectColor = Color.cyan;
            transform.SetParent(parentGO.transform, false);

            ResizeRecTransform(popupGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject dropDownButtonGO = transform.Find("Button").gameObject;
            ReformatUnityButton(dropDownButtonGO, width * (1 - labelProportion), (height * .9f) - sliderHeight, fontSize, -2f, 1.5f, AnchorPresets.bottomRight, true);

            GameObject textGO = transform.Find("Text").gameObject;
            ReformatText(textGO, width * (labelProportion), height * .9f, fontSize, 0f, 0f, AnchorPresets.middleLeft);

            GameObject popupPanelGO = transform.Find("PopupPanel").gameObject;
            GameObject scrollViewGO = popupPanelGO.transform.Find("Scroll View").gameObject;

            ScrollRect scrollRect = scrollViewGO.GetComponent<ScrollRect>();
            scrollRect.verticalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;

            if (popupAbove) ReformatPopupPanel(popupPanelGO, width * (1 - labelProportion), popupHeight, fontSize, 0f, height, anchorPreset: AnchorPresets.bottomRight);
            else ReformatPopupPanel(popupPanelGO, width * (1 - labelProportion), popupHeight, fontSize, 0f, -height, anchorPreset: AnchorPresets.topRight);

            GameObject sliderGO = transform.Find("Slider").gameObject;
            ReformatScrollPopupSlider(sliderGO, width * (1 - labelProportion), sliderHeight, 0f, 0f, anchorPreset: AnchorPresets.topRight, clearSprite: false);

            return uiPopup;
        }


        public static void RecursiveTrans(Transform transform, string spaces)
        {

            foreach (Transform childTrans in transform)
            {
                SuperController.LogMessage(spaces + childTrans.name);
                string newSpaces = spaces + "    ";
                RecursiveTrans(childTrans, newSpaces);
            }
        }

        public static UIDynamicPopup CreateFilterablePopup(GameObject parentGO, JSONStorableStringChooser jssc, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, float popupHeight = 300f, int anchorPreset = AnchorPresets.none)
        {
            float labelProportion = 0.3f;
            Transform transform = GameObject.Instantiate<Transform>(Globals.mvrScript.manager.configurableFilterablePopupPrefab);

            GameObject popupGO = transform.gameObject;

            UIDynamicPopup uiPopup = popupGO.transform.GetComponent<UIDynamicPopup>();
            uiPopup.label = label;
            uiPopup.popup.selectColor = Color.cyan;
            jssc.RegisterPopup(uiPopup.popup);

            transform.SetParent(parentGO.transform, false);

            ResizeRecTransform(popupGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject dropDownButtonGO = transform.Find("Button").gameObject;
            ReformatUnityButton(dropDownButtonGO, width * (1 - labelProportion), height * 0.6f, fontSize, 0f, 0f, AnchorPresets.bottomRight, false);

            GameObject labelTextGO = transform.Find("Text").gameObject;
            ReformatText(labelTextGO, width * labelProportion, height, fontSize, 0f, 0f, AnchorPresets.middleLeft);

            GameObject inputFieldGO = transform.Find("InputField").gameObject;
            ReformatInputField(inputFieldGO, width * (1 - labelProportion), height * 0.4f, fontSize, 0f, 0f, AnchorPresets.topRight);

            GameObject popupPanelGO = transform.Find("PopupPanel").gameObject;
            ReformatPopupPanel(popupPanelGO, width * (1 - labelProportion), popupHeight, fontSize, 0f, -height, anchorPreset: AnchorPresets.topRight);
            GameObject scrollViewGO = popupPanelGO.transform.Find("Scroll View").gameObject;

            ScrollRect scrollRect = scrollViewGO.GetComponent<ScrollRect>();
            scrollRect.verticalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;
            return uiPopup;
        }
        private static void ReformatPopupPanel(GameObject popupPanelGO, float width, float height, int fontSize = 12, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none)
        {
            float scrollBarWidth = 10f;
            ResizeRecTransform(popupPanelGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject buttonTemplateGO = popupPanelGO.transform.Find("PopupValueButtonTemplate").gameObject;
            ReformatUnityButton(buttonTemplateGO, width - scrollBarWidth, 15f, fontSize, clearSprite: true, anchorPreset: AnchorPresets.none);

            if (popupPanelGO.transform.Find("Scroll View") != null)
            {
                ReformatScrollView(popupPanelGO.transform.Find("Scroll View").gameObject, width, scrollBarWidth, height);
            }
        }

        private static void ReformatScrollView(GameObject scrollViewGO, float width, float vScrollBarWidth, float height = 25, float hScrollBarHeight = 0)
        {
            GameObject viewPortGO = scrollViewGO.transform.Find("Viewport").gameObject;
            //            ResizeRecTransform(viewPortGO, width-vScrollBarWidth, height- hScrollBarHeight);

            RectTransform viewPortRT = viewPortGO.transform.GetComponent<RectTransform>();
            viewPortRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height - hScrollBarHeight / 2);
            viewPortRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width - vScrollBarWidth / 2);

            if (scrollViewGO.transform.Find("Scrollbar Vertical") != null)
            {
                GameObject scrollBarVerticalGO = scrollViewGO.transform.Find("Scrollbar Vertical").gameObject;
                Scrollbar sb = scrollBarVerticalGO.transform.GetComponent<Scrollbar>();
                sb.enabled = vScrollBarWidth > 0f;
                
                RectTransform scrollBarRT = scrollBarVerticalGO.transform.GetComponent<RectTransform>();
                scrollBarRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, vScrollBarWidth);
                scrollBarRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            }
            if (scrollViewGO.transform.Find("Scrollbar Horizontal") != null)
            {
                GameObject scrollBarVerticalGO = scrollViewGO.transform.Find("Scrollbar Horizontal").gameObject;
                Scrollbar sb = scrollBarVerticalGO.transform.GetComponent<Scrollbar>();
                sb.enabled = hScrollBarHeight > 0f;
                RectTransform scrollBarRT = scrollBarVerticalGO.transform.GetComponent<RectTransform>();
                scrollBarRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, hScrollBarHeight);
                scrollBarRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width  );

            }


        }

        private static void ReformatScrollPopupSlider(GameObject sliderGO, float width, float height, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool clearSprite = true)
        {
            ResizeRecTransform(sliderGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject handleGO = sliderGO.transform.Find("Handle Slide Area/Handle").gameObject;
            Image handleImage = handleGO.transform.GetComponent<Image>();
            if (clearSprite) handleImage.sprite = null;
        }

        private static void ReformatImage(GameObject imageGO, float width, float height, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool clearSprite = true)
        {
            ResizeRecTransform(imageGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);
            Image popupImage = imageGO.transform.GetComponent<Image>();
            if (clearSprite) popupImage.sprite = null;
            popupImage.color = new Color(0.9f, 0.9f, 0.9f);
        }
        private static void ReformatInputField(GameObject inputFieldGO, float width, float height, int fontSize = 12, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none)
        {
            Image popupImage = inputFieldGO.transform.GetComponent<Image>();
            popupImage.sprite = null;

            ResizeRecTransform(inputFieldGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject buttonGO = inputFieldGO.transform.Find("Button").gameObject;
            ReformatUnityButton(buttonGO, 15f, height, fontSize - 4, -40f, clearSprite: false, textAlignment: TextAnchor.MiddleLeft);

            GameObject countTextGO = inputFieldGO.transform.Find("CountText").gameObject;
            ReformatText(countTextGO, 40f, height, fontSize - 4, anchorPreset: AnchorPresets.middleRight);

            GameObject countTextBGGO = inputFieldGO.transform.Find("CountTextBG").gameObject;
            ReformatImage(countTextBGGO, 40f, height, anchorPreset: AnchorPresets.middleRight, clearSprite: true);
            GameObject maxCountBGGO = inputFieldGO.transform.Find("MaxIndicator").gameObject;
            ReformatImage(maxCountBGGO, 40f, height, anchorPreset: AnchorPresets.middleRight, clearSprite: true);

            GameObject inputTextGO = inputFieldGO.transform.Find("Text").gameObject;
            ReformatText(inputTextGO, width - 55f, height, fontSize, 0f, 0f, AnchorPresets.middleLeft, textAlginment: TextAnchor.MiddleLeft);

            GameObject placeholderTextGO = inputFieldGO.transform.Find("Placeholder").gameObject;
            ReformatText(placeholderTextGO, width - 55f, height, fontSize, 0f, 0f, AnchorPresets.middleLeft, textAlginment: TextAnchor.MiddleLeft);
        }

        private static void ReformatText(GameObject textGO, float width, float height, int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, TextAnchor textAlginment = TextAnchor.MiddleCenter, bool resizeTextForBestFit = false)
        {

            RectTransform textRT = textGO.transform.GetComponent<RectTransform>();
            textRT.localScale = new Vector3(0.5f, 0.5f, 0.5f);
            width = width * 2;
            height = height * 2;

            ResizeRecTransform(textGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            Text text = textGO.transform.GetComponent<Text>();
            text.alignment = textAlginment;
            text.fontSize = fontSize;
            text.resizeTextForBestFit = resizeTextForBestFit;
        }
        private static void ReformatUnityButton(GameObject buttonGO, float width, float height, int fontSize = 12, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool clearSprite = true, TextAnchor textAlignment = TextAnchor.MiddleCenter)
        {
            ResizeRecTransform(buttonGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            if (clearSprite)
            {
                Image buttonImage = buttonGO.transform.GetComponent<Image>();
                buttonImage.sprite = null;
            }

            Transform textTransform = buttonGO.transform.Find("Text");
            ReformatText(textTransform.gameObject, width - 4f, height, fontSize, 2f, 0f, anchorPreset: AnchorPresets.centre, textAlginment: textAlignment);
        }

        public static void ResizeRecTransform(GameObject go, float width, float height, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none)
        {
            RectTransform rt = go.GetComponent<RectTransform>();
            ResizeRecTransform(rt, width, height, anchorXOffset, anchorYOffset, anchorPreset);

        }

        public static void ResizeRecTransform(RectTransform rt, float width, float height, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none)
        {
            rt.sizeDelta = Vector2.zero;
            if (anchorPreset != AnchorPresets.none)
            {
                
                rt.anchorMin = AnchorPresets.GetAnchorMin(anchorPreset);
                rt.anchorMax = AnchorPresets.GetAnchorMax(anchorPreset);
                rt.pivot = AnchorPresets.GetPivot(anchorPreset);
            }
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            rt.anchoredPosition = new Vector2(anchorXOffset, anchorYOffset);
        }

        public static void ResizeWidth(GameObject go, float width)
        {
            RectTransform rt = go.GetComponent<RectTransform>();
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
        }

        public static void ResizeHeight(GameObject go, float height)
        {
            RectTransform rt = go.GetComponent<RectTransform>();
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
        }

        public static void ResetBottomOffset(GameObject go, float offset)
        {
            RectTransform rt = go.GetComponent<RectTransform>();
            rt.offsetMin = new Vector2(rt.offsetMin.x, offset);
        }

        public static void ResetAnchorPosition(GameObject go, float xPosition, float yPosition)
        {
            RectTransform rt = go.GetComponent<RectTransform>();
            rt.anchoredPosition = new Vector2(xPosition, yPosition);
        }
        public static void ResetAnchorType(GameObject go, int anchorPreset)
        {
            RectTransform rt = go.GetComponent<RectTransform>();
            rt.anchorMin = AnchorPresets.GetAnchorMin(anchorPreset);
            rt.anchorMax = AnchorPresets.GetAnchorMax(anchorPreset);
            rt.pivot = AnchorPresets.GetPivot(anchorPreset);
        }

    }

    public class TriStateToggle:UIDynamic
    {
        private JSONStorableBool toggleJSB;

        public UIDynamicToggle toggleUID { get; private set; }

        public Toggle toggle { get { return toggleUID.toggle; } }

        public Color backgroundColor { get { return toggleUID.backgroundColor; } set { toggleUID.backgroundColor = value; } }
        public Color textColor { get { return toggleUID.textColor; } set { toggleUID.textColor = value; } }

        public string label { get { return toggleUID.label; } set { toggleUID.label = value; } }

        public Text labelText { get { return toggleUID.labelText; } set { toggleUID.labelText = value; } }

        public new float height { get { return toggleUID.height; } set { toggleUID.height = value; } }

        public TriStateToggle (JSONStorableBool _toggleJSB, UIDynamicToggle _toggleUID)
        {
            toggleJSB = _toggleJSB;
            toggleUID = _toggleUID;
        }

        public void RefreshToggle(int triState)
        {
            var backgroundTrans = toggle.transform.Find("Background");
            var image = backgroundTrans.gameObject.GetComponent<Image>();

            if (triState != TriState.inactive) image.color = Color.cyan;
            else image.color = Color.white;

            if (triState == TriState.active) toggleJSB.valNoCallback = true;
            else toggleJSB.valNoCallback = false;
        }

    }

    public static class AnchorPresets
    {
        public const int none = 0;
        public const int topLeft = 1;
        public const int topMiddle = 2;
        public const int topRight = 3;
        public const int middleLeft = 4;
        public const int centre = 5;
        public const int middleRight = 6;
        public const int bottomLeft = 7;
        public const int bottomMiddle = 8;
        public const int bottomRight = 9;

        public const int vStretchMiddle = 11;
        public const int hStretchMiddle = 12;
        public const int vStretchRight = 13;
        public const int vStretchLeft = 14;
        public const int hStretchTop = 15;
        public const int vhStretch = 16;

        public static Vector2 GetAnchorMin(int anchorPreset)
        {
            Vector2 anchorMin = new Vector2();

            if (anchorPreset == topLeft) anchorMin = new Vector2(0f, 1f);
            if (anchorPreset == topMiddle) anchorMin = new Vector2(0.5f, 1f);
            if (anchorPreset == topRight) anchorMin = new Vector2(1f, 1f);
            if (anchorPreset == middleLeft) anchorMin = new Vector2(0f, 0.5f);
            if (anchorPreset == centre) anchorMin = new Vector2(0.5f, 0.5f);
            if (anchorPreset == middleRight) anchorMin = new Vector2(1f, 0.5f);
            if (anchorPreset == bottomLeft) anchorMin = new Vector2(0f, 0f);
            if (anchorPreset == bottomMiddle) anchorMin = new Vector2(0.5f, 0f);
            if (anchorPreset == bottomRight) anchorMin = new Vector2(1f, 0f);

            if (anchorPreset == vStretchMiddle) anchorMin = new Vector2(0.5f, 0f);
            if (anchorPreset == vStretchRight) anchorMin = new Vector2(1f, 0f);
            if (anchorPreset == vStretchLeft) anchorMin = new Vector2(0f, 0f);
            if (anchorPreset == hStretchMiddle) anchorMin = new Vector2(0f, 0.5f);
            if (anchorPreset == hStretchTop) anchorMin = new Vector2(0f, 1f);
            if (anchorPreset == vhStretch) anchorMin = new Vector2(0f, 0f);

            return anchorMin;
        }

        public static Vector2 GetAnchorMax(int anchorPreset)
        {
            Vector2 anchorMax = new Vector2();

            if (anchorPreset == topLeft) anchorMax = new Vector2(0f, 1f);
            if (anchorPreset == topMiddle) anchorMax = new Vector2(0.5f, 1f);
            if (anchorPreset == topRight) anchorMax = new Vector2(1f, 1f);
            if (anchorPreset == middleLeft) anchorMax = new Vector2(0f, 0.5f);
            if (anchorPreset == centre) anchorMax = new Vector2(0.5f, 0.5f);
            if (anchorPreset == middleRight) anchorMax = new Vector2(1f, 0.5f);
            if (anchorPreset == bottomLeft) anchorMax = new Vector2(0f, 0f);
            if (anchorPreset == bottomMiddle) anchorMax = new Vector2(0.5f, 0f);
            if (anchorPreset == bottomRight) anchorMax = new Vector2(1f, 0f);

            if (anchorPreset == vStretchMiddle) anchorMax = new Vector2(0.5f, 1f);
            if (anchorPreset == vStretchRight) anchorMax = new Vector2(1f, 1f);
            if (anchorPreset == vStretchLeft) anchorMax = new Vector2(0f, 1f);
            if (anchorPreset == hStretchMiddle) anchorMax = new Vector2(1f, 0.5f);
            if (anchorPreset == hStretchTop) anchorMax = new Vector2(1f, 1f);
            if (anchorPreset == vhStretch) anchorMax = new Vector2(1f, 1f);

            return anchorMax;
        }

        public static Vector2 GetPivot(int anchorPreset)
        {
            Vector2 pivot = new Vector2();

            if (anchorPreset == topLeft) pivot = new Vector2(0f, 1f);
            if (anchorPreset == topMiddle) pivot = new Vector2(0.5f, 1f);
            if (anchorPreset == topRight) pivot = new Vector2(1f, 1f);
            if (anchorPreset == middleLeft) pivot = new Vector2(0f, 0.5f);
            if (anchorPreset == centre) pivot = new Vector2(0.5f, 0.5f);
            if (anchorPreset == middleRight) pivot = new Vector2(1f, 0.5f);
            if (anchorPreset == bottomLeft) pivot = new Vector2(0f, 0f);
            if (anchorPreset == bottomMiddle) pivot = new Vector2(0.5f, 0f);
            if (anchorPreset == bottomRight) pivot = new Vector2(1f, 0f);

            if (anchorPreset == vStretchMiddle) pivot = new Vector2(0.5f, 1f);
            if (anchorPreset == hStretchMiddle) pivot = new Vector2(0f, 0.5f);
            if (anchorPreset == vStretchRight) pivot = new Vector2(1f, 1f);
            if (anchorPreset == vStretchLeft) pivot = new Vector2(0f, 1f);
            if (anchorPreset == hStretchTop) pivot = new Vector2(0f, 1f);
            if (anchorPreset == vhStretch) pivot = new Vector2(0f, 1f);
            return pivot;
        }

    }


    public class PrimaryActionPointerEnterExitHelp : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        private ResourceTypeSelections resourceTypeSelections;
        private JSONStorableString helpTextJSS;
        private ResourcePrimaryActionUI resourcePrimaryActionUI;
        public void Init(JSONStorableString _helpTextJSS, ResourceTypeSelections _resourceTypeSelections, ResourcePrimaryActionUI _resourcePrimaryActionUI)
        {
            helpTextJSS = _helpTextJSS;
            resourceTypeSelections = _resourceTypeSelections;
            resourcePrimaryActionUI = _resourcePrimaryActionUI;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {
            ResourceVersionGroupEntry rvge = resourceTypeSelections.GetFirstSelectedRVGE();
            if (rvge.resourceType == ResourceType.scene) helpTextJSS.val = "Load Scene: '" + rvge.resourceDisplayName + "'";
            else if (rvge.resourceCategory == ResourceCategory.preset)
            {
                if (resourcePrimaryActionUI.targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag) helpTextJSS.val = "Load Preset'" + rvge.resourceDisplayName + "' to Atom '" + resourcePrimaryActionUI.targetAtomJSMultiEnum.mainVal + "'";
                else helpTextJSS.val = "Create a target atom to apply preset '" + rvge.resourceDisplayName + "'";
            }
            else if (rvge.resourceCategory == ResourceCategory.plugins) helpTextJSS.val = "Load Plugin: '" + rvge.resourceDisplayName + "'";

        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpTextJSS.val = "";
        }
    }

    public class FavPointerEnterExitHelp : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        private ResourceTypeSelections resourceSelections;
        private BrowserFiltersUI browserFiltersUI;
        private JSONStorableString helpTextJSS;
        public void Init(JSONStorableString _helpTextJSS, ResourceTypeSelections _resourceSelections, BrowserFiltersUI _browserFiltersUI)
        {
            helpTextJSS = _helpTextJSS;
            resourceSelections = _resourceSelections;
            browserFiltersUI = _browserFiltersUI;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {
            helpTextJSS.val = "Set selected Resource(s) as Favourites.";
            if (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi || resourceSelections.isFavourite == TriState.active) helpTextJSS.val = "Unset selected Resource(s) as Favourites.";
            else if (resourceSelections.isFavourite == TriState.indeterminate)
            {
                int selectedResourceCount = resourceSelections.GetSelectedCount();
                int notFavouriteResourcesCount = selectedResourceCount - resourceSelections.GetFavouriteCount();
                helpTextJSS.val = helpTextJSS.val + " WARNING: The Favourite state of " + notFavouriteResourcesCount.ToString() + " resource(s) out of " + selectedResourceCount.ToString() + " selected resources will be changed. This can't be easily reversed without manually identifying which resource(s) were not previously set as Favourites.";

                Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
                image.color = Color.red;
            }
        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpTextJSS.val = "";

            Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
            image.color = Color.grey;
        }
    }
    public class HiddenPointerEnterExitHelp : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        private ResourceTypeSelections resourceSelections;
        private BrowserFiltersUI browserFiltersUI;
        private JSONStorableString helpTextJSS;
        public void Init(JSONStorableString _helpTextJSS, ResourceTypeSelections _resourceSelections, BrowserFiltersUI _browserFiltersUI)
        {
            helpTextJSS = _helpTextJSS;
            resourceSelections = _resourceSelections;
            browserFiltersUI = _browserFiltersUI;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {
            helpTextJSS.val = "Set selected Resource(s) as Hidden.";
            if (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi || resourceSelections.isHidden == TriState.active) helpTextJSS.val = "Unset selected Resource(s) as Hidden.";
            else if (resourceSelections.isHidden == TriState.indeterminate)
            {
                int selectedResourceCount = resourceSelections.GetSelectedCount();
                int notHiddenResourcesCount = selectedResourceCount - resourceSelections.GetHiddenCount();
                helpTextJSS.val = helpTextJSS.val + " WARNING: The Hidden state of " + notHiddenResourcesCount.ToString() + " resource(s) out of " + selectedResourceCount.ToString() + " selected resources will be changed. This can't be easily reversed without manually identifying which resource(s) were not previously set as Hidden.";

                Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
                image.color = Color.red;
            }
        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpTextJSS.val = "";

            Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
            image.color = Color.grey;
        }
    }

    public class TagPointerEnterExitHelp : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        private Tag baTag;
        private ResourceTypeSelections resourceSelections;
        private BrowserFiltersUI browserFiltersUI;
        private JSONStorableString helpTextJSS;
        public void Init(Tag _tag, JSONStorableString _helpTextJSS, ResourceTypeSelections _resourceSelections, BrowserFiltersUI _browserFiltersUI)
        {
            helpTextJSS = _helpTextJSS;
            baTag = _tag;
            resourceSelections = _resourceSelections;
            browserFiltersUI = _browserFiltersUI;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {
            helpTextJSS.val = "Set Tag '" + baTag.tagName + "' for the selected Resources.";
            if (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi || resourceSelections.IsTagged(baTag) == TriState.active) helpTextJSS.val = "Unset Tag '" + baTag.tagName + "' for the selected Resources.";
            else if (resourceSelections.IsTagged(baTag) == TriState.indeterminate)
            {
                int selectedResourceCount = resourceSelections.GetSelectedCount();
                int notTaggedResourcesCount = selectedResourceCount - resourceSelections.GetTaggedCount(baTag);
                helpTextJSS.val = helpTextJSS.val + " WARNING: The tagged state of " + notTaggedResourcesCount.ToString() + " resource(s) out of " + selectedResourceCount.ToString() + " selected resources will be changed. This can't be easily reversed without manually identifying which resource(s) were not previously tagged.";

                Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
                image.color = Color.red;
            }
        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpTextJSS.val = "";

            Image image = helpTextJSS.text.transform.parent.gameObject.GetComponent<Image>();
            image.color = Color.grey;
        }
    }

    public class GeneralPointerEnterExitHelp : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        private string helpText;
        private JSONStorableString helpTextJSS;
        public void Init(string _helpText, JSONStorableString _helpTextJSS)
        {
            helpTextJSS = _helpTextJSS;
            helpText = _helpText;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {

            helpTextJSS.val = helpText;
        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpTextJSS.val = "";
        }
    }

    public class ResourceButtonPointerEnterExit : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        private UIDynamicButton button;
        private Dictionary<UIDynamicButton, ResourceVersionGroupEntry> rvgeDict;
        private JSONStorableString helpText;
        public void Init(UIDynamicButton _button, Dictionary<UIDynamicButton, ResourceVersionGroupEntry> _rvgeDict, JSONStorableString _helpText)
        {
            button = _button;
            rvgeDict = _rvgeDict;
            helpText = _helpText;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {
            if (rvgeDict.ContainsKey(button))
            {
                helpText.val = rvgeDict[button].resourceFullPathReference;
            }

        }

        public void OnPointerExit(PointerEventData eventData)
        {
            if (rvgeDict.ContainsKey(button))
            {
                helpText.val = "";
            }
        }
    }
    public class SelectedResourceButtonPointerEnterExit : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
    {
        private ResourceVersionGroupEntry rvge;
        private JSONStorableString helpText;
        public void Init(ResourceVersionGroupEntry _rvge, JSONStorableString _helpText)
        {
            rvge = _rvge;
            helpText = _helpText;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {
            helpText.val = rvge.resourceFullPathReference;

        }

        public void OnPointerExit(PointerEventData eventData)
        {
            helpText.val = "";
        }

    }


    public class ButtonPointerDrag : MonoBehaviour, IBeginDragHandler, IDragHandler
    {
        private Canvas mainCanvas;
        private Dictionary<UIDynamicButton, ResourceVersionGroupEntry> rvgeDict;
        private JSONStorableString helpText;

        private float distance;


        public void Init(Canvas _mainCanvas)
        {
            mainCanvas = _mainCanvas;
        }
        public void OnBeginDrag(PointerEventData eventData)
        {
            distance = Vector3.Distance(transform.position, SuperController.singleton.leftControllerCamera.transform.position);
        }

        public void OnDrag(PointerEventData eventData)
        {
            Vector3 buttonToCanvasOffset = transform.position - mainCanvas.transform.position;

            Ray leftControllerRay = new Ray(SuperController.singleton.leftControllerCamera.transform.position, SuperController.singleton.leftControllerCamera.transform.forward);

            Vector3 rayPoint = leftControllerRay.GetPoint(distance);
            //Move the GameObject when you drag it
            mainCanvas.transform.position = rayPoint - buttonToCanvasOffset;
        }
    }

}
