﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.UI;
using UnityEngine.XR;


namespace JayJayWon
{

    public class BrowserUI
    {
        public static Color darkBlueGreyTransparent = new Color(0.4f, 0.4f, 0.6f, 0.98f);
        public static Color lightGrey = new Color(0.9f, 0.9f, 0.9f, 1f);
        public static Color mediumGrey = new Color(0.5f, 0.5f, 0.5f, 1f);
        public static Color darkGrey = new Color(0.2f, 0.2f, 0.2f, 1f);

        private GameObject moreOptionsGameObject;
        private Image fileBrowserImage;
        private Image hubBrowserImage;
        private CanvasScaler mainCanvasScalar;
        public BrowserUI()
        {
            if (XRSettings.enabled) isVREnabled = true;
            else isVREnabled = false;

            GameObject canvasObject = new GameObject();
            mainCanvas = canvasObject.AddComponent<Canvas>();
            mainCanvas.pixelPerfect = false;

            if (isVREnabled) mainCanvas.renderMode = RenderMode.WorldSpace;
            else mainCanvas.renderMode = RenderMode.ScreenSpaceOverlay;

            mainCanvasScalar = canvasObject.AddComponent<CanvasScaler>();
            GraphicRaycaster rayCaster = canvasObject.AddComponent<GraphicRaycaster>();

            GameObject canvasLaunchButtonObject = new GameObject();
            launchButtonCanvas = canvasLaunchButtonObject.AddComponent<Canvas>();
            launchButtonCanvas.renderMode = RenderMode.WorldSpace;
            launchButtonCanvas.pixelPerfect = false;

            canvasLaunchButtonObject.layer = SuperController.singleton.mainHUD.gameObject.layer;
            launchButtonCanvas.transform.SetParent(SuperController.singleton.mainHUD, false);
            CanvasScaler launchButtonCS = canvasLaunchButtonObject.AddComponent<CanvasScaler>();
            launchButtonCS.scaleFactor = 100f;
            launchButtonCS.dynamicPixelsPerUnit = 1f;
            GraphicRaycaster gr = canvasLaunchButtonObject.AddComponent<GraphicRaycaster>();

            mainCanvas.sortingOrder = -3;

            vamUILaunchButton = UI.CreateUIButton(launchButtonCanvas.gameObject, 38f, 36f, "Browser Assist", 20);
            vamUILaunchButton.buttonColor = Color.white;
            string thumbnailPath = Utils.GetLatestVARPath("JayJayWon.BrowserAssist.latest:/Custom/Scripts/JayJayWon/BrowserAssist/BrowserAssist.jpg");
            ImageUtils.QueueLoadTexture(thumbnailPath, LaunchButtonImageLoadedCallback, true);
            vamUILaunchButton.button.onClick.AddListener(delegate
            {
                ToggleActive();
            });

            SuperController.singleton.AddCanvas(launchButtonCanvas);
            SuperController.singleton.AddCanvas(mainCanvas);

            launchButtonCanvas.transform.localScale = new Vector3(0.001f, 0.001f, 0.001f);
            launchButtonCanvas.transform.localPosition = new Vector3(-0.131f, -0.074f, 0f);
            if (isVREnabled)
            {
                launchButtonCanvas.transform.localEulerAngles = new Vector3(32, 180, 0);
                launchButtonCanvas.transform.Translate(new Vector3(0f, -0.01f, -0.039f));
                mainCanvasScalar.dynamicPixelsPerUnit = 4f;

                mainCanvas.transform.localScale = new Vector3(Const._localCanvasScale, Const._localCanvasScale, Const._localCanvasScale);

            }
            else
            {
                launchButtonCanvas.transform.localEulerAngles = new Vector3(0, 180, 0);
                mainCanvasScalar.dynamicPixelsPerUnit = 2f;
                mainCanvasScalar.uiScaleMode = CanvasScaler.ScaleMode.ScaleWithScreenSize;
                RefreshScale();

            }
            browserFiltersUI = new BrowserFiltersUI(this);
            resourceSelectionsUI = new ResourceSelections(this);

            helpTextJSS = new JSONStorableString("helpText", "");
            
            InitBrowserUI();

            var moreOptionstransform = SuperController.singleton.mainHUD.Find("LowerHUDPivot/LowerHUDFlip/Scene Control Canvas/Panel/ClickForMoreOptions/OtherOptionsPopup");

            moreOptionsGameObject = moreOptionstransform.gameObject;


            var fileBrowserTransform = SuperController.singleton.worldUI.Find("Canvas/FileBrowserWorld/MainPanel");
            var hubBrowserTransform = SuperController.singleton.worldUI.Find("Canvas/HubBrowsePanel");
            fileBrowserImage = fileBrowserTransform.GetComponent<Image>();
            hubBrowserImage = hubBrowserTransform.GetComponent<Image>();

            SetInactive();
        }

        public void RefreshScale()
        {
            if (!isVREnabled)  mainCanvasScalar.referenceResolution = new Vector2(1440 / Settings.uiScaleJSF.val, 900 / Settings.uiScaleJSF.val);
        }

        private void LaunchButtonImageLoadedCallback(ImageLoaderThreaded.QueuedImage qi)
        {
            if (qi.tex != null)
            {
                Texture2D tex = qi.tex;
                vamUILaunchButton.button.image.sprite = Sprite.Create(tex, new Rect(0.0f, 0.0f, tex.width, tex.height), new Vector2(0.5f, 0.5f), 100.0f);
                vamUILaunchButton.buttonText.text = "";
            }
        }

        const int filterBoxHeight = 50;
        const int boxSpacerHeight = 3;
        public static int edgeSpacerSize = 2;
        const int infoTextBoxHeight = 22;
        public const int resourceActionBoxHeight = 90;



        private int screenHeight { get { if (isVREnabled) return 600;
                else return Screen.height; } }

        private int resourceBrowserBottomOffset { get
            {
                return edgeSpacerSize + infoTextBoxHeight + resourceActionBoxHeight + resourceSelectionsUI.resourceSelectionBoxHeight + (boxSpacerHeight * 3);
            }
        }
        private int resourceBrowserTopOffset
        {
            get
            {
                return edgeSpacerSize + filterBoxHeight + boxSpacerHeight;
            }
        }

        private int resourceBrowserHeight { get
            {
                return screenHeight - resourceBrowserBottomOffset - resourceBrowserTopOffset;
            } }

        public int selectedResourcesVOffset
        {
            get
            {
                return edgeSpacerSize + infoTextBoxHeight + resourceActionBoxHeight + (boxSpacerHeight * 2);
            }
        }

        public float browserScreenWidth { get
            {
                if (isVREnabled) return 650f;
                if (currentDesktopDisplayMode == DesktopDisplayMode.rightDocked || currentDesktopDisplayMode == DesktopDisplayMode.leftDocked) return 650f;
                if (currentDesktopDisplayMode == DesktopDisplayMode.fullWidth) return 1440 / Settings.uiScaleJSF.val*0.9f;
                    return -1f;
            } }

        public int backgroundBoxAnchorPreset { get
            {
                if (isVREnabled) return AnchorPresets.centre;
                if (currentDesktopDisplayMode == DesktopDisplayMode.rightDocked) return AnchorPresets.vStretchRight;
                if (currentDesktopDisplayMode == DesktopDisplayMode.fullWidth) return AnchorPresets.vStretchMiddle;
                if (currentDesktopDisplayMode == DesktopDisplayMode.leftDocked) return AnchorPresets.vStretchLeft;
                return -1;
            } }

        private void InitBrowserUI()
        {
            float vSize = 0f;
            if (isVREnabled) vSize = screenHeight;

            backgroundBoxGO = UI.AddChildImage(parentTransform: mainCanvas.transform,
                color: new Color(0.5f, 0.5f, 0.5f, 0.95f),
                anchorPreset: backgroundBoxAnchorPreset,
                horizontalSize: browserScreenWidth,
                verticalSize: vSize,
                anchoredPositionOffset: new Vector2(-5f, 0f));
           
            resourceBrowserGO = UI.CreateVScrollableContent(backgroundBoxGO, new Color(0.8f, 0.8f, 0.8f, 0.8f), AnchorPresets.vStretchMiddle, browserScreenWidth-10f, resourceBrowserHeight, new Vector2(0f, -filterBoxHeight- edgeSpacerSize- boxSpacerHeight));


           selectedResourceGO = UI.AddChildGOImage(parentGO:backgroundBoxGO,
                color: darkBlueGreyTransparent,
                anchorPreset: AnchorPresets.bottomMiddle,
                horizontalSize: resourceSelectionsUI.resourceSelectionBoxWidth,
                verticalSize: resourceSelectionsUI.resourceSelectionBoxHeight,
                anchoredPositionOffset: new Vector2((resourceSelectionsUI.resourceSelectionBoxWidth -640f)/2, selectedResourcesVOffset));

            resourceActionsGO = UI.AddChildGOImage(parentGO: backgroundBoxGO,
                color: darkBlueGreyTransparent,
                anchorPreset: AnchorPresets.bottomMiddle,
                horizontalSize: 640f,
                verticalSize: resourceActionBoxHeight,
                anchoredPositionOffset: new Vector2(0f, infoTextBoxHeight+edgeSpacerSize+boxSpacerHeight));

            filtersBoxGO = UI.AddChildGOImage(parentGO: backgroundBoxGO,
                color: darkBlueGreyTransparent,
                anchorPreset: AnchorPresets.topMiddle,
                horizontalSize: 640f,
                verticalSize: filterBoxHeight,
                anchoredPositionOffset: new Vector2(0f, -edgeSpacerSize));

            cancelButtonUID = UI.CreateUIButton(backgroundBoxGO, 20f, 20f, "<b>X</b>", 28, (650f / 2) - 15f, edgeSpacerSize+1f, anchorPreset: AnchorPresets.bottomMiddle) ;
            cancelButtonUID.buttonImage.sprite = null;
            cancelButtonUID.button.onClick.AddListener(delegate {
                SetInactive();
            });
            ApplyHelp(cancelButtonUID, "Close the Browser Assist window.");

            

            if (isVREnabled)
            {
                repositionUI = UI.CreateUIButton(backgroundBoxGO, 25f, 25f, "<b>O</b>", 34, 1, 2, anchorPreset: AnchorPresets.bottomLeft);
                var component = repositionUI.button.gameObject.AddComponent<ButtonPointerDrag>();
                component.Init(mainCanvas);
            }
            else
            {
                leftExpandContractButtonUID = UI.CreateUIButton(backgroundBoxGO, 20f, 20f, "<b>|<</b>", 28, -(640f / 2) + 10f, edgeSpacerSize+1f, anchorPreset: AnchorPresets.bottomMiddle);
                leftExpandContractButtonUID.buttonImage.sprite = null;
                leftExpandContractButtonUID.button.onClick.AddListener(delegate {
                    if (currentDesktopDisplayMode == DesktopDisplayMode.leftDocked)
                    {
                        currentDesktopDisplayMode = DesktopDisplayMode.fullWidth;
                        leftExpandContractButtonUID.label = "<b>|<</b>";
                        ApplyHelp(leftExpandContractButtonUID, "Dock the Browser Assist window on the left of the screen.");
                    }
                    else
                    {
                        currentDesktopDisplayMode = DesktopDisplayMode.leftDocked;
                        leftExpandContractButtonUID.label = "<b>>></b>";
                        
                        ApplyHelp(leftExpandContractButtonUID, "Expand the Browser Assist window across the screen width.");
                    }
                    rightExpandContractButtonUID.label = "<b>>|</b>";
                    ApplyHelp(rightExpandContractButtonUID, "Dock the Browser Assist window on the right of the screen.");
                    RefreshBrowserUIDimensions();
                    RepositionBrowserBottomUIElements();
                });
                rightExpandContractButtonUID = UI.CreateUIButton(backgroundBoxGO, 20f, 20f, "<b><<</b>", 28, -(640f / 2) + 32f, edgeSpacerSize+1f, anchorPreset: AnchorPresets.bottomMiddle);
                rightExpandContractButtonUID.buttonImage.sprite = null;
                rightExpandContractButtonUID.button.onClick.AddListener(delegate {
                    if (currentDesktopDisplayMode == DesktopDisplayMode.rightDocked)
                    {
                        currentDesktopDisplayMode = DesktopDisplayMode.fullWidth;
                        
                        rightExpandContractButtonUID.label = "<b>>|</b>";
                        ApplyHelp(rightExpandContractButtonUID, "Dock the Browser Assist window on the right of the screen.");

                    }
                    else
                    {
                        currentDesktopDisplayMode = DesktopDisplayMode.rightDocked;
                        rightExpandContractButtonUID.label = "<b><<</b>";
                        ApplyHelp(rightExpandContractButtonUID, "Expand the Browser Assist window across the screen width.");

                    }
                    leftExpandContractButtonUID.label = "<b>|<</b>";
                    ApplyHelp(leftExpandContractButtonUID, "Dock the Browser Assist window on the left of the screen.");
                    RefreshBrowserUIDimensions();
                    RepositionBrowserBottomUIElements();
                });
            }
            infoTextBoxGO = UI.CreateBasicTextBox(backgroundBoxGO, helpTextJSS, 574f, infoTextBoxHeight, 22, 12f, edgeSpacerSize, AnchorPresets.bottomMiddle, true, 18);          

            
            browserFiltersUI.InitUI(filtersBoxGO);
            resourceSelectionsUI.InitUI(selectedResourceGO);
            resourceActionsUI = new ResourceActionsUI(this,resourceActionsGO);
            resourceActionsUI.RefreshUI();

            resourceBrowserUI = new ResourceBrowserUI(this, resourceBrowserGO);
        }

        private void RepositionBrowserBottomUIElements()
        {
            if (currentDesktopDisplayMode == DesktopDisplayMode.leftDocked)
            {
                UI.ResetAnchorPosition(rightExpandContractButtonUID.gameObject, (640f / 2) - 10f, edgeSpacerSize + 1f);
                UI.ResetAnchorPosition(leftExpandContractButtonUID.gameObject, (640f / 2) - 32f, edgeSpacerSize + 1f);
                UI.ResetAnchorPosition(cancelButtonUID.gameObject, -(650f / 2) + 15f, edgeSpacerSize + 1f);

                UI.ResetAnchorPosition(infoTextBoxGO, -12f, edgeSpacerSize);
            }
            else
            {
                UI.ResetAnchorPosition(rightExpandContractButtonUID.gameObject, -(640f / 2) +32f, edgeSpacerSize + 1f);
                UI.ResetAnchorPosition(leftExpandContractButtonUID.gameObject, -(640f / 2) + 10f, edgeSpacerSize + 1f);
                UI.ResetAnchorPosition(cancelButtonUID.gameObject, (650f / 2) - 15f, edgeSpacerSize + 1f);
                UI.ResetAnchorPosition(infoTextBoxGO, 12f, edgeSpacerSize);
            }
        }
        private void RefreshBrowserUIDimensions()
        {
            UI.ResizeWidth(backgroundBoxGO, browserScreenWidth);
            UI.ResetAnchorType(backgroundBoxGO, backgroundBoxAnchorPreset);
            UI.ResizeWidth(resourceBrowserGO, browserScreenWidth - 10f);
            GameObject buttonListViewportGO = resourceBrowserGO.transform.Find("scrollableContentViewportGO").gameObject;
            UI.ResizeWidth(buttonListViewportGO, browserScreenWidth - 20f);
            resourceBrowserUI.RepopulateButtons();
        }
        public void RefreshBrowserUILayout()
        {       
            UI.ResizeHeight(selectedResourceGO, resourceSelectionsUI.resourceSelectionBoxHeight);
            UI.ResizeWidth(selectedResourceGO, resourceSelectionsUI.resourceSelectionBoxWidth);
            UI.ResetAnchorPosition(selectedResourceGO, (resourceSelectionsUI.resourceSelectionBoxWidth - 640f) / 2, selectedResourcesVOffset);
            resourceBrowserUI.RepositionNavButtons();

            UI.ResetBottomOffset(resourceBrowserGO, (edgeSpacerSize * 1) + infoTextBoxHeight + resourceActionBoxHeight + resourceSelectionsUI.resourceSelectionBoxHeight + (boxSpacerHeight * 3));
        }

        public void ToggleActive()
        {
            if (mainCanvas.enabled) SetInactive();
            else SetActive();
        }
        public void SetActive()
        {
            RefreshFilterResults();

            if (isVREnabled)
            {
                Vector3 headPosition = SuperController.singleton.lookCamera.transform.position;
                if (SuperController.singleton.mainHUD == null || !SuperController.singleton.mainHUD.gameObject.activeSelf)
                {
                    
                    Quaternion headFacingDirection = SuperController.singleton.lookCamera.transform.rotation;
                    Vector3 offsetPosition = new Vector3(0f, 0f, 2f);

                    Vector3 newPosition = headPosition + headFacingDirection * offsetPosition;
                    mainCanvas.transform.position = new Vector3(newPosition.x, headPosition.y, newPosition.z);
                }
                else
                {
                    Vector3 newTargetPos = SuperController.singleton.mainHUD.position + (SuperController.singleton.navigationRig.up * 1f) - (SuperController.singleton.mainHUD.forward * 0.6f);

                    Vector3 newPos = new Vector3(newTargetPos.x, headPosition.y, newTargetPos.z);
                    mainCanvas.transform.position = newPos;
                    
                    // SuperController.singleton.activeUI = SuperController.ActiveUI.None;

                    SuperController.singleton.HideMainHUD();
                }
            }            
            mainCanvas.enabled = true;
        }
        public void SetInactive()
        {
            mainCanvas.enabled = false;
//            if (isVREnabled && SuperController.singleton.activeUI==SuperController.ActiveUI.None) SuperController.singleton.activeUI = SuperController.singleton.lastActiveUI;
                
        }
        public bool isUIActive { get { return mainCanvas.enabled; } }
        public void RefreshButtonLayout(bool forceSelectedResourceRefresh=false)
        {
            currentFilterResultsPage = 1;
            if (resourceBrowserUI!=null) resourceBrowserUI.RepopulateButtons(forceSelectedResourceRefresh);
        }

        public void RefreshFilterResults(bool forceSelectedResourceRefresh = false)
        {
            currentFilterResults = ResourceManifest.GetResourceGroups(browserFiltersUI);
            RefreshButtonLayout(forceSelectedResourceRefresh);
        }

        public void Update()
        {
            resourceBrowserUI.Update();
            if (isVREnabled)
            {
                Vector3 headPosition = SuperController.singleton.lookCamera.transform.position;
                mainCanvas.transform.LookAt(headPosition, SuperController.singleton.navigationRig.up);
                mainCanvas.transform.rotation *= Quaternion.Euler(0f, 180f,0f);
                
            }

            if (moreOptionsGameObject.activeSelf) vamUILaunchButton.gameObject.SetActive(false);
            else vamUILaunchButton.gameObject.SetActive(true);

            if (fileBrowserImage.gameObject.activeInHierarchy || hubBrowserImage.gameObject.activeInHierarchy)
            {
                if (!isHidden && mainCanvas!=null && mainCanvas.enabled)
                {
                    SetInactive();
                    isHidden = true;
                }
                
            }
            else
            {
                if (isHidden)
                {
                    SetActive();
                    isHidden = false;
                }
            }
        }

        public void OnDestroy()
        {
            if (SuperController.singleton != null)
            {
                SuperController.singleton.RemoveCanvas(mainCanvas);
                SuperController.singleton.RemoveCanvas(launchButtonCanvas);
            }
            if (mainCanvas != null) mainCanvas.transform.SetParent(null, false);
            if (launchButtonCanvas != null) launchButtonCanvas.transform.SetParent(null, false);

            if (mainCanvas.gameObject != null) GameObject.Destroy(mainCanvas.gameObject);
            if (launchButtonCanvas.gameObject != null) GameObject.Destroy(launchButtonCanvas.gameObject);
        }

        public void ApplyHelp(UIDynamicPopup popupUID, string helpText)
        {
            ApplyHelp(popupUID.popup.gameObject, helpText);
            ApplyHelp(popupUID.popup.popupPanel.gameObject, helpText);
        }
        public void ApplyHelp(UIDynamicToggle toggleUID, string helpText)
        {
            ApplyHelp(toggleUID.toggle.gameObject, helpText);
        }
        public void ApplyHelp(UIDynamicTextField textFieldUID, string helpText)
        {
            GameObject gameObject = textFieldUID.UItext.gameObject;
            if (textFieldUID.transform.GetComponent<InputField>() != null) gameObject = textFieldUID.transform.GetComponent<InputField>().gameObject;
            ApplyHelp(gameObject, helpText);
        }
        public void ApplyHelp(UIDynamicButton buttonUID, string helpText)
        {
            ApplyHelp(buttonUID.button.gameObject, helpText);
        }

        public void ApplyHelp(GameObject gameObject, string helpText)
        {
            GeneralPointerEnterExitHelp contextHelp = gameObject.AddComponent<GeneralPointerEnterExitHelp>();
            contextHelp.Init(helpText, helpTextJSS);
        }

        public void ApplyTagHelp(UIDynamicToggle toggleUID, string helpText)
        {
            ApplyHelp(toggleUID.toggle.gameObject, helpText);
        }


        public int currentDesktopDisplayMode { get; private set; } = DesktopDisplayMode.rightDocked;

        private GameObject backgroundBoxGO;
        private GameObject filtersBoxGO;
        private GameObject resourceBrowserGO;
        private GameObject selectedResourceGO;
        private GameObject resourceActionsGO;
        private GameObject actionBoxGO;
        private GameObject helpGO;

        private UIDynamicButton vamUILaunchButton;

        private UIDynamicButton repositionUI;
        private UIDynamicButton leftExpandContractButtonUID;
        private UIDynamicButton rightExpandContractButtonUID;
        private UIDynamicButton cancelButtonUID;
        private GameObject infoTextBoxGO;

        public JSONStorableString helpTextJSS;
 

        public BrowserFiltersUI browserFiltersUI;
        public ResourceActionsUI resourceActionsUI;
        public ResourceBrowserUI resourceBrowserUI;

        public List<ResourceVersionGroupEntry> currentFilterResults;
        public int currentFilterResultsPage = 1;

        public ResourceSelections resourceSelectionsUI ;

        private Canvas mainCanvas;
        private Canvas launchButtonCanvas;
        public bool isVREnabled { get; private set; }

        private bool isHidden = false;
    }

    public class ResourceSelections
    {
        private Dictionary<int, ResourceTypeSelections> resourceTypeSelection = new Dictionary<int, ResourceTypeSelections>();

        private GameObject resourceSelectionsScrollableBoxGO;
        private GameObject resourceSelectionsBoxGO;
        private GameObject resourceSelectionScrollableViewportContentGO;

        private BrowserFiltersUI browserFiltersUI;
        private BrowserUI browserUI;

        private List<GameObject> resourceSelectionRows = new List<GameObject>();

        const int resourceSelectionBoxHeightMin = 20;
        const int resourceSelectionBoxHeightMax = 55;
        const int resourceSelectionBoxWidthMin = 456;
        const int resourceSelectionBoxWidthMax = 545;

        public ResourceTypeSelections GetResourceTypeSelections(int resourceType)
        {
            if (resourceTypeSelection.ContainsKey(resourceType)) return resourceTypeSelection[resourceType];
            return null;
        }

        public int resourceSelectionBoxHeight
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return resourceSelectionBoxHeightMin;
                return resourceSelectionBoxHeightMax;
            }
        }
        public int resourceSelectionBoxWidth
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return resourceSelectionBoxWidthMin;
                return resourceSelectionBoxWidthMax;
            }
        }

        public ResourceSelections(BrowserUI _browserUI)
        {
            browserUI = _browserUI;
            browserFiltersUI = _browserUI.browserFiltersUI;
            
        }
        public void InitUI(GameObject _resourceSelectionsBoxGO)
        {
            resourceSelectionsBoxGO = _resourceSelectionsBoxGO;
            resourceSelectionsScrollableBoxGO = UI.CreateVScrollableContent(resourceSelectionsBoxGO, BrowserUI.darkBlueGreyTransparent, AnchorPresets.vhStretch, resourceSelectionBoxWidth, resourceSelectionBoxHeight, new Vector2(0f, 0f));
            resourceSelectionScrollableViewportContentGO = resourceSelectionsScrollableBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;

            RePopulateSelectionsBox();

        }

        private void ClearResourceSelectionRows()
        {
            foreach (var rowGO in resourceSelectionRows) GameObject.Destroy(rowGO);
            resourceSelectionRows.Clear();
        }
        public void RePopulateSelectionsBox()
        {
            int resourceType = browserFiltersUI.resourceTypeJSEnum.val;
            ClearResourceSelectionRows();

            if (browserUI!=null && browserUI.resourceBrowserUI!=null) browserUI.resourceBrowserUI.SetMultiSelectClear();

            if (resourceTypeSelection.ContainsKey(resourceType))
            {

                GameObject currentRowGO = CreateSelectResourceRow();

                float currentRowResourcesWidth = 0f;

                List<ResourceVersionGroupEntry> selectedRVGEs;
                if (browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi) selectedRVGEs = resourceTypeSelection[resourceType].selectedRVGEs;
                else 
                {
                    selectedRVGEs = new List<ResourceVersionGroupEntry>();
                    var firstSelectedRVGE = resourceTypeSelection[resourceType].GetFirstSelectedRVGE();
                    if (firstSelectedRVGE!=null) selectedRVGEs.Add(firstSelectedRVGE);
                }

                foreach (var resourceVersionGroupEntry in selectedRVGEs)
                {
                    float nextButtonWidth = GetRVGEButtonWidth(resourceVersionGroupEntry);
                    float cancelButtonWidth = 0f;
                    if (selectedRVGEs.Count > 1 && browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi) cancelButtonWidth = 18f;

                    if (currentRowResourcesWidth + nextButtonWidth + cancelButtonWidth > resourceSelectionBoxWidth)
                    {
                        currentRowGO = CreateSelectResourceRow();
                        currentRowResourcesWidth = 0f;
                    }
                    var resourceButton = UI.CreateUIButton(currentRowGO, nextButtonWidth, 16f, resourceVersionGroupEntry.resourceDisplayName, 20, currentRowResourcesWidth, 0, AnchorPresets.middleLeft);
                    resourceButton.buttonImage.sprite = null;
                    resourceButton.button.interactable = false;
                    resourceButton.buttonText.resizeTextForBestFit = true;
                    resourceButton.buttonText.resizeTextMinSize = 15;
                    resourceButton.buttonText.resizeTextMaxSize = 21;

                    if (resourceVersionGroupEntry == selectedRVGEs.First()) resourceButton.buttonColor = Color.cyan;
                    ColorBlock colors = resourceButton.button.colors;
                    colors.disabledColor = BrowserUI.lightGrey;
                    resourceButton.button.colors = colors ;
                    SelectedResourceButtonPointerEnterExit click = resourceButton.button.gameObject.AddComponent<SelectedResourceButtonPointerEnterExit>();
                    click.Init(resourceVersionGroupEntry, browserUI.helpTextJSS);

                    currentRowResourcesWidth += nextButtonWidth ;

                    if (cancelButtonWidth>0f)
                    {
                        var cancelButton = UI.CreateUIButton(currentRowGO, 16f, 16f, "X", 22, currentRowResourcesWidth, 0, AnchorPresets.middleLeft);
                        cancelButton.buttonImage.sprite = null;
                        cancelButton.buttonColor = Color.cyan;
                        currentRowResourcesWidth += 18f;
                        SelectedResourceButtonPointerEnterExit click2 = cancelButton.button.gameObject.AddComponent<SelectedResourceButtonPointerEnterExit>();
                        click2.Init(resourceVersionGroupEntry, browserUI.helpTextJSS);
                        cancelButton.button.onClick.AddListener(delegate
                        {
                            DeSelectRVGE(resourceVersionGroupEntry);
                        });
                    }
                    currentRowResourcesWidth += 3f;
                }
                if (selectedRVGEs.Count < 2 && !browserUI.isVREnabled && browserUI.browserFiltersUI.selectModeJSEnum.val==BrowserSelectMode.multi)
                {
                    var guideTextJSS = new JSONStorableString("guideText", "<i>Hold <Ctrl> key to add or remove a clicked resource to these selections. Use left <Shift> to select multiple contiguous resources prior to the clicked resource. Use right <Shift> to select multiple contiguous resources after the clicked resource. Use <Alt> key to change the last selected resource.</i>");
                    var textBox = UI.CreateBasicTextBox(currentRowGO, guideTextJSS, resourceSelectionBoxWidth- currentRowResourcesWidth, 40, 20, currentRowResourcesWidth, 0f, AnchorPresets.topLeft,  backgoundColor: new Color(0f,0f,0f,0f));

                }
            }
            if (browserUI.resourceActionsUI!=null) browserUI.resourceActionsUI.RefreshUI();
        }

        private GameObject CreateSelectResourceRow()
        {
            GameObject currentRowGO = UI.AddChildGOImage(resourceSelectionScrollableViewportContentGO, new Color(0f, 0.8f, 0f, 0f), AnchorPresets.hStretchTop, resourceSelectionBoxWidth, 20f, Vector2.zero);
            resourceSelectionRows.Add(currentRowGO);

            return currentRowGO;
        }

        private float GetRVGEButtonWidth(ResourceVersionGroupEntry rvge)
        {
            if (browserFiltersUI.selectModeJSEnum.val!=BrowserSelectMode.multi) return Mathf.Clamp(rvge.resourceDisplayName.Length * 7, 80f, 400f);
            return Mathf.Clamp(rvge.resourceDisplayName.Length * 6,40f,200f);

        }

        public int SelectionCount(int resourceType)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return 0;
            return resourceTypeSelection[resourceType].selectedRVGEs.Count;
        }
        public bool IsSelected(ResourceVersionGroupEntry rvge, int selectionMode, UIDynamicButton updatedButton)
        {
            if (!resourceTypeSelection.ContainsKey(rvge.resourceType)) resourceTypeSelection.Add(rvge.resourceType, new ResourceTypeSelections(browserUI.resourceBrowserUI));
            return resourceTypeSelection[rvge.resourceType].IsSelected(rvge, selectionMode, updatedButton);
        }

        public void SingleSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button)
        {
            if (!resourceTypeSelection.ContainsKey(rvge.resourceType)) resourceTypeSelection.Add(rvge.resourceType, new ResourceTypeSelections(browserUI.resourceBrowserUI));
            resourceTypeSelection[rvge.resourceType].SingleSelectRVGE(rvge, button);

            RePopulateSelectionsBox();
        }

        public void LeftShiftSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button, List<ResourceVersionGroupEntry> currentPageRVGEs, List<UIDynamicButton> currentPageButtons)
        {
            if (!resourceTypeSelection.ContainsKey(rvge.resourceType)) resourceTypeSelection.Add(rvge.resourceType, new ResourceTypeSelections(browserUI.resourceBrowserUI));
            resourceTypeSelection[rvge.resourceType].LeftShiftSelectRVGE(rvge, button, currentPageRVGEs, currentPageButtons);

            RePopulateSelectionsBox();
        }
        public void RightShiftSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button, List<ResourceVersionGroupEntry> currentPageRVGEs, List<UIDynamicButton> currentPageButtons)
        {
            if (!resourceTypeSelection.ContainsKey(rvge.resourceType)) resourceTypeSelection.Add(rvge.resourceType, new ResourceTypeSelections(browserUI.resourceBrowserUI));
            resourceTypeSelection[rvge.resourceType].RightShiftSelectRVGE(rvge, button, currentPageRVGEs, currentPageButtons);

            RePopulateSelectionsBox();
        }

        public void ToggleSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button)
        {
            if (!resourceTypeSelection.ContainsKey(rvge.resourceType)) resourceTypeSelection.Add(rvge.resourceType, new ResourceTypeSelections(browserUI.resourceBrowserUI));
            resourceTypeSelection[rvge.resourceType].ToggleSelectRVGE(rvge, button);

            RePopulateSelectionsBox();
        }
        public void ClearSelectedRVGEButtons(int resourceType)
        {
            if (resourceTypeSelection.ContainsKey(resourceType)) resourceTypeSelection[resourceType].ClearSelectedRVGEButtons();
        }

        public void DeSelectRVGE(ResourceVersionGroupEntry rvge)
        {
            resourceTypeSelection[rvge.resourceType].DeSelectRVGE(rvge);

            RePopulateSelectionsBox();
        }

        public void DeSelectAll(int resourceType, bool keepFirst = true)
        {
            if (resourceTypeSelection.ContainsKey(resourceType)) resourceTypeSelection[resourceType].DeSelectAll(keepFirst);

            RePopulateSelectionsBox();
        }

        public void SelectAll(int resourceType, List<ResourceVersionGroupEntry> currentFilterResults)
        {
            if (resourceTypeSelection.ContainsKey(resourceType)) resourceTypeSelection[resourceType].SelectAll(currentFilterResults);

            RePopulateSelectionsBox();
        }

        public void SwitchSelectMode(int resourceType, int selectionMode)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) resourceTypeSelection.Add(resourceType, new ResourceTypeSelections(browserUI.resourceBrowserUI));
            resourceTypeSelection[resourceType].SwitchSelectMode(selectionMode);

            RePopulateSelectionsBox();
        }

        public ResourceVersionGroupEntry GetFirstSelectedRVGE(int resourceType)
        {
            if (resourceTypeSelection.ContainsKey(resourceType)) return resourceTypeSelection[resourceType].GetFirstSelectedRVGE();
            return null;
        }

        public int IsSelectionTagged(int resourceType, int selectMode, Tag tag)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return TriState.inactive;
            if (selectMode != BrowserSelectMode.multi) return resourceTypeSelection[resourceType].IsFirstTagged(tag);
            return resourceTypeSelection[resourceType].IsTagged(tag);
        }

        public int IsSelectionFavourite(int resourceType, int selectMode)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return TriState.inactive;
            if (selectMode != BrowserSelectMode.multi) return resourceTypeSelection[resourceType].isFirstFavourite;
            return resourceTypeSelection[resourceType].isFavourite;
        }

        public int IsSelectionHidden(int resourceType, int selectMode)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return TriState.inactive;
            if (selectMode != BrowserSelectMode.multi) return resourceTypeSelection[resourceType].isFirstHidden;
            return resourceTypeSelection[resourceType].isHidden;
        }
        public void SetSelectionTag(int resourceType, int selectMode, Tag tag, int tagTriState)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return;
            if (selectMode != BrowserSelectMode.multi) resourceTypeSelection[resourceType].SetFirstTagged(tag, tagTriState);
            else resourceTypeSelection[resourceType].SetTagged(tag, tagTriState);
        }
        public void SetSelectionFavourite (int resourceType, int selectMode, int favTriState)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return ;
            if (selectMode != BrowserSelectMode.multi) resourceTypeSelection[resourceType].isFirstFavourite = favTriState;
            else resourceTypeSelection[resourceType].isFavourite = favTriState;
        }

        public void SetSelectionHidden(int resourceType, int selectMode, int hiddenTriState)
        {
            if (!resourceTypeSelection.ContainsKey(resourceType)) return;
            if (selectMode != BrowserSelectMode.multi) resourceTypeSelection[resourceType].isFirstHidden = hiddenTriState;
            else resourceTypeSelection[resourceType].isHidden = hiddenTriState;
        }
    }

    public class ResourceTypeSelections
    {
        private Dictionary<ResourceVersionGroupEntry, UIDynamicButton> selectedRVGEButtons = new Dictionary<ResourceVersionGroupEntry, UIDynamicButton>();

        public List<ResourceVersionGroupEntry> selectedRVGEs = new List<ResourceVersionGroupEntry>();

        private ResourceBrowserUI resourceBrowserUI;

        public int isHidden { get
            {
                bool hiddenFound = false;
                bool notHiddenFound = false;
                foreach (var rvge in selectedRVGEs)
                {
                    if (rvge.isHidden) hiddenFound = true;
                    else notHiddenFound = true;

                    if (hiddenFound && notHiddenFound) return TriState.indeterminate;
                }

                if (hiddenFound) return TriState.active;
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool hide = (value == TriState.active);
                foreach (var rvge in selectedRVGEs)
                {
                    rvge.isHidden = hide;
                }
            }
        }

        public int isFirstHidden
        {
            get
            {
                
                if (selectedRVGEs.Count > 0 && GetFirstSelectedRVGE().isHidden) return TriState.active;
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool hide = (value == TriState.active);
                GetFirstSelectedRVGE().isHidden = hide;
            }
        }

        public int isFavourite
        {
            get
            {
                bool favFound = false;
                bool notFavFound = false;
                foreach (var rvge in selectedRVGEs)
                {
                    if (rvge.isFavourite) favFound = true;
                    else notFavFound = true;

                    if (favFound && notFavFound) return TriState.indeterminate;
                }

                if (favFound) return TriState.active;
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool fav = (value == TriState.active);
                foreach (var rvge in selectedRVGEs)
                {
                    rvge.isFavourite = fav;
                }
            }
        }

        public int isFirstFavourite
        {
            get
            {
                if (selectedRVGEs.Count > 0 && GetFirstSelectedRVGE().isFavourite) return TriState.active;
                return TriState.inactive;
            }
            set
            {
                if (value == TriState.indeterminate) return;
                bool fav = (value == TriState.active);
                GetFirstSelectedRVGE().isFavourite = fav;
            }
        }


        public ResourceTypeSelections(ResourceBrowserUI _resourceBrowserUI)
        {
            resourceBrowserUI = _resourceBrowserUI;
        }
        public void SwitchSelectMode(int selectMode)
        {
            if (selectedRVGEs.Count > 0)
            {
                UIDynamicButton firstButton = null;

                if (selectedRVGEButtons.ContainsKey(selectedRVGEs.First()))
                    firstButton = selectedRVGEButtons[selectedRVGEs.First()];
                foreach (var rvgeButton in selectedRVGEButtons)
                {
                    if ( selectMode == BrowserSelectMode.multi || rvgeButton.Value == firstButton ) ResourceBrowserUI.SelectColor(rvgeButton.Value);
                    else ResourceBrowserUI.DeSelectColor(rvgeButton.Value);
                }

            }

        }

        public void ClearSelectedRVGEButtons()
        {
            selectedRVGEButtons.Clear();
        }

        public void RightShiftSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button, List<ResourceVersionGroupEntry> currentPageRVGEs, List<UIDynamicButton> currentPageButtons)
        {
            int selectionIndex = currentPageButtons.FindIndex(x => x == button);
            int altIndex = currentPageButtons.Count-1;

            bool selectToggle = !selectedRVGEs.Contains(rvge);

            for (int searchIndex = selectionIndex +1; searchIndex < currentPageButtons.Count; searchIndex++)
            {
                if (selectToggle && selectedRVGEs.Contains(currentPageRVGEs[searchIndex]))
                {
                    altIndex = searchIndex -1;
                    break;
                }
                if (!selectToggle && !selectedRVGEs.Contains(currentPageRVGEs[searchIndex]))
                {
                    altIndex = searchIndex -1;
                    break;
                }
            }

            for (int toggleIndex = altIndex; toggleIndex >= selectionIndex; toggleIndex--)
            {
                ToggleSelectRVGE(currentPageRVGEs[toggleIndex], currentPageButtons[toggleIndex]);
            }

        }

        public void LeftShiftSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button, List<ResourceVersionGroupEntry> currentPageRVGEs, List<UIDynamicButton> currentPageButtons)
        {
            int selectionIndex = currentPageButtons.FindIndex(x => x==button);
            int altIndex = 0;

            bool selectToggle = !selectedRVGEs.Contains(rvge);

            for (int searchIndex = selectionIndex - 1; searchIndex >= 0; searchIndex--)
            {
                if (selectToggle && selectedRVGEs.Contains(currentPageRVGEs[searchIndex]))
                {
                    altIndex = searchIndex + 1;
                    break;
                }
                if (!selectToggle && !selectedRVGEs.Contains(currentPageRVGEs[searchIndex]))
                {
                    altIndex = searchIndex + 1;
                    break;
                }
            }
    
            for (int toggleIndex = altIndex; toggleIndex <= selectionIndex; toggleIndex++)
            {
                ToggleSelectRVGE(currentPageRVGEs[toggleIndex], currentPageButtons[toggleIndex]);
            }

        }
        public void ToggleSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button)
        {
            if (!selectedRVGEs.Contains(rvge)) SelectRVGE(rvge, button);
            else DeSelectRVGE(rvge);
        }
        private void SelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button)
        {
            if (selectedRVGEs.Contains(rvge)) selectedRVGEs.Remove(rvge);
            selectedRVGEs.Insert(0,rvge);

            if (!selectedRVGEButtons.ContainsKey(rvge)) selectedRVGEButtons.Add(rvge, button);
            else selectedRVGEButtons[rvge] = button;
            if (button!=null) ResourceBrowserUI.SelectColor(button);
        }
        public void SingleSelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton button)
        {
            if (selectedRVGEs.Count > 0)
            {
                ResourceVersionGroupEntry firstRVGE = selectedRVGEs.First();
                if (selectedRVGEButtons.ContainsKey(firstRVGE)) ResourceBrowserUI.DeSelectColor(selectedRVGEButtons[firstRVGE]);

               selectedRVGEs.Remove(firstRVGE);
                selectedRVGEButtons.Remove(firstRVGE);
            }
            SelectRVGE(rvge, button);

        }
        public void DeSelectRVGE(ResourceVersionGroupEntry rvge)
        {
            
            selectedRVGEs.Remove(rvge);
            if (selectedRVGEButtons.ContainsKey(rvge))
            {
                UIDynamicButton button = selectedRVGEButtons[rvge];
                selectedRVGEButtons.Remove(rvge);
                if (button!=null) ResourceBrowserUI.DeSelectColor(button);
            }
                
        }

        public ResourceVersionGroupEntry GetFirstSelectedRVGE()
        {
            if (selectedRVGEs.Count>0) return selectedRVGEs.First();
            return null;
        }

        public void DeSelectAll(bool keepFirst)
        {
            ResourceVersionGroupEntry firstRVGE = null;

            if (selectedRVGEs.Count>0) firstRVGE = selectedRVGEs.First();

            UIDynamicButton firstRVGEButton =null;
            foreach (var kvp in selectedRVGEButtons)
            {
                if (kvp.Key != firstRVGE)
                {
                   if (kvp.Value != null) ResourceBrowserUI.DeSelectColor(kvp.Value);
                } 
                else firstRVGEButton = kvp.Value;
            }
            selectedRVGEs.Clear();
            selectedRVGEButtons.Clear();
            if (firstRVGE!=null && keepFirst)
            {
                selectedRVGEs.Add(firstRVGE);
                selectedRVGEButtons.Add(firstRVGE, firstRVGEButton);
            }
            
        }

        public void SelectAll(List<ResourceVersionGroupEntry> currentFilterResults)
        {
            foreach (var rvge in currentFilterResults)
            {
                SelectRVGE(rvge, resourceBrowserUI.GetUIDButton(rvge));
            }
        }

        public bool IsSelected(ResourceVersionGroupEntry rvge, int selectionMode, UIDynamicButton updatedButton)
        {
            if (!selectedRVGEs.Contains(rvge)) return false;
            if (!selectedRVGEButtons.ContainsKey(rvge)) selectedRVGEButtons.Add(rvge, updatedButton);
            else selectedRVGEButtons[rvge] = updatedButton;

            ResourceVersionGroupEntry firstRVGE = selectedRVGEs.First();
            if (rvge == firstRVGE || selectionMode == BrowserSelectMode.multi) return true;
            return false;
        }

        public int IsFirstTagged(Tag tag)
        {
            if (selectedRVGEs.Count > 0 && GetFirstSelectedRVGE().IsTagged(tag)) return TriState.active;
            return TriState.inactive;
        }

        public int IsTagged(Tag tag)
        {
            bool tagFound = false;
            bool notTagFound = false;
            foreach (var rvge in selectedRVGEs)
            {
                if (rvge.IsTagged(tag)) tagFound = true;
                else notTagFound = true;

                if (tagFound && notTagFound) return TriState.indeterminate;
            }

            if (tagFound) return TriState.active;
            return TriState.inactive;
        }
        public void SetTagged(Tag tag, int value)
        {
            if (value == TriState.indeterminate) return;
            bool tagValue = (value == TriState.active);
            foreach (var rvge in selectedRVGEs)
            {
                rvge.SetTag(tag, tagValue);
            }
        }
        public int GetFavouriteCount()
        {
            int count = 0;
            foreach (var rvge in selectedRVGEs)
            {
                if (rvge.isFavourite) count++;
            }

            return count;
        }
        public int GetHiddenCount()
        {
            int count = 0;
            foreach (var rvge in selectedRVGEs)
            {
                if (rvge.isHidden) count++;
            }

            return count;
        }
        public int GetTaggedCount(Tag tag)
        {
            int count = 0;
            foreach (var rvge in selectedRVGEs)
            {
                if (rvge.IsTagged(tag)) count++;
            }

            return count;
        }

        public int GetSelectedCount()
        {
            return selectedRVGEs.Count();
        }

        public void SetFirstTagged(Tag tag, int value)
        {
            if (value == TriState.indeterminate) return;
            bool tagValue = (value == TriState.active);
            GetFirstSelectedRVGE().SetTag(tag, tagValue);
        }

    }
    public class ResourceBrowserUI
    {
        private GameObject parentBoxGO;
        private BrowserUI parentBrowser;
        private BrowserFiltersUI browserFiltersUI;

        public JSONStorableString resourceCountTextJSS;

        private int resourceButtonSize = 4;
        const int resourceButtonlabelHeight = 16;

        private UIDynamicButton firstPageUIDButton;
        private UIDynamicButton prevPageUIDButton;
        private UIDynamicButton nextPageUIDButton;
        private UIDynamicButton incResourceButtonSizesUIDButton;
        private UIDynamicButton decResourceButtonSizesUIDButton;
        private UIDynamicButton clearMultiSelectResourcesUIDButton;
        private UIDynamicButton selectAllResourcesUIDButton;
        private GameObject resourceCountTextBox;

        private Dictionary<UIDynamicButton, ResourceVersionGroupEntry> UIDButtonToRVGEDict = new Dictionary<UIDynamicButton, ResourceVersionGroupEntry>();
        private Dictionary<ResourceVersionGroupEntry, UIDynamicButton> RVGEToUIDButtonDict = new Dictionary<ResourceVersionGroupEntry, UIDynamicButton>();

        private List<GameObject> activeResourceRowGOs = new List<GameObject>();
        private List<GameObject> inactiveResourceRowGOs = new List<GameObject>();

        private List<UIDynamicButton> activeButtonUIDs = new List<UIDynamicButton>();
        private HashSet<UIDynamicButton> unparentedButtonUIDs = new HashSet<UIDynamicButton>();

        private Dictionary<GameObject, List<UIDynamicButton>> buttonUIDsByRow = new Dictionary<GameObject, List<UIDynamicButton>>();
        private List<ResourceVersionGroupEntry> currentFilterResults { get { return parentBrowser.currentFilterResults; } }
        private int currentFilterResultsPage { get { return parentBrowser.currentFilterResultsPage; } set { parentBrowser.currentFilterResultsPage = value; } }

        private List<ResourceVersionGroupEntry> currentPageRVGEs = new List<ResourceVersionGroupEntry>();
        private List<UIDynamicButton> currentPageButtons = new List<UIDynamicButton>();
        private float resouceButtonWidth
        {
            get
            {
                if (resourceButtonSize == 0) return 60;
                if (resourceButtonSize == 1) return 67;
                if (resourceButtonSize == 2) return 75.5f;
                if (resourceButtonSize == 3) return 87;
                if (resourceButtonSize == 4) return 102;
                if (resourceButtonSize == 5) return 123;
                if (resourceButtonSize == 6) return 154;
                return 123;
            }
        }
        private int buttonsPerRow
        {
            get
            {
                return (int)Math.Floor((parentBrowser.browserScreenWidth-20f) / (float)(resouceButtonWidth + 3));
            }
        }
        private int selectedResourcesVOffset
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return -23;
                else return -45;
            }
        }
        
        private float navButtonsHOffset
        {
            get
            {
                return parentBrowser.resourceSelectionsUI.resourceSelectionBoxWidth - (640/2) + BrowserUI.edgeSpacerSize +7.5f ;
            }
        }
        private float navButtonsVOffset
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return selectedResourcesVOffset;
                else return selectedResourcesVOffset + 5;
            }
        }
        private float resourceCountHOffset
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return navButtonsHOffset + 47f+(76f/2);
                else return navButtonsHOffset + (76f / 2);
            }
        }
        private int resourceCountVOffset
        {
            get
            {
                if (browserFiltersUI == null || browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi) return selectedResourcesVOffset;
                else return selectedResourcesVOffset + 25;
            }
        }
        private int resoureButtonSizeHOffset
        {
            get
            {
                return (640/2)-30;
            }
        }
        public ResourceBrowserUI(BrowserUI _parentBrowser, GameObject _parentBoxGO)
        {
            parentBrowser = _parentBrowser;
            parentBoxGO = _parentBoxGO;
            browserFiltersUI = parentBrowser.browserFiltersUI;

            firstPageUIDButton = UI.CreateUIButton(parentBoxGO, 15f, 20f, "<b>|<</b>", 26, navButtonsHOffset, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            firstPageUIDButton.buttonImage.sprite = null;
            firstPageUIDButton.button.onClick.AddListener(delegate {
                currentFilterResultsPage = 1;
                RepopulateButtons();
            });
            parentBrowser.ApplyHelp(firstPageUIDButton, "Navigate to the first page of browser results");

            prevPageUIDButton = UI.CreateUIButton(parentBoxGO, 15f, 20f, "<b><</b>", 26, navButtonsHOffset + 15, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            prevPageUIDButton.buttonImage.sprite = null;
            prevPageUIDButton.button.onClick.AddListener(delegate {
                currentFilterResultsPage--;
                RepopulateButtons();
            });
            parentBrowser.ApplyHelp(prevPageUIDButton, "Navigate to the previous page of browser results");

            nextPageUIDButton = UI.CreateUIButton(parentBoxGO, 15f, 20f, "<b>></b>", 26, navButtonsHOffset + 30, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            nextPageUIDButton.buttonImage.sprite = null;
            nextPageUIDButton.button.onClick.AddListener(delegate {
                if ((currentFilterResultsPage) * Settings.maxButtonsPerPageJSF.val < currentFilterResults.Count)
                {
                    currentFilterResultsPage++;
                    RepopulateButtons();
                }

            });
            parentBrowser.ApplyHelp(nextPageUIDButton, "Navigate to the next page of browser results");

            incResourceButtonSizesUIDButton = UI.CreateUIButton(parentBoxGO, 20f, 20f, "<b>+</b>", 26, resoureButtonSizeHOffset, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            incResourceButtonSizesUIDButton.buttonImage.sprite = null;
            incResourceButtonSizesUIDButton.button.onClick.AddListener(delegate {
                if (resourceButtonSize < 6) resourceButtonSize++;
                RepopulateButtons();
            });
            parentBrowser.ApplyHelp(incResourceButtonSizesUIDButton, "Increase the size of resource icons in the browser results window.");

            decResourceButtonSizesUIDButton = UI.CreateUIButton(parentBoxGO, 20f, 20f, "<b>-</b>", 26, resoureButtonSizeHOffset+20, selectedResourcesVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            decResourceButtonSizesUIDButton.buttonImage.sprite = null;
            decResourceButtonSizesUIDButton.button.onClick.AddListener(delegate {
                if (resourceButtonSize > 0) resourceButtonSize--;
                RepopulateButtons();
            });
            parentBrowser.ApplyHelp(decResourceButtonSizesUIDButton, "Decrease the size of resource icons in the browser results window.");

            clearMultiSelectResourcesUIDButton = UI.CreateUIButton(parentBoxGO, 15f, 15f, "X", 22, navButtonsHOffset, selectedResourcesVOffset-31f, anchorPreset: AnchorPresets.bottomMiddle);
            clearMultiSelectResourcesUIDButton.buttonImage.sprite = null;
            clearMultiSelectResourcesUIDButton.buttonColor = Color.cyan;
            clearMultiSelectResourcesUIDButton.button.onClick.AddListener(delegate {
                parentBrowser.resourceSelectionsUI.DeSelectAll(browserFiltersUI.resourceTypeJSEnum.val);
            });
            clearMultiSelectResourcesUIDButton.gameObject.SetActive(false);
            parentBrowser.ApplyHelp(clearMultiSelectResourcesUIDButton, "Clear all resources (other than the very latest selected resource) from the selection window.");

            selectAllResourcesUIDButton = UI.CreateUIButton(parentBoxGO,65f, 15f, "Select All", 22, navButtonsHOffset, navButtonsVOffset, anchorPreset: AnchorPresets.bottomMiddle);
            selectAllResourcesUIDButton.buttonImage.sprite = null;
            selectAllResourcesUIDButton.buttonColor = Color.cyan;
            selectAllResourcesUIDButton.button.onClick.AddListener(delegate {

                if (parentBrowser.isVREnabled || (!Input.GetKey(KeyCode.LeftControl)) && !Input.GetKey(KeyCode.RightControl)) parentBrowser.resourceSelectionsUI.DeSelectAll(browserFiltersUI.resourceTypeJSEnum.val,false);
                parentBrowser.resourceSelectionsUI.SelectAll(browserFiltersUI.resourceTypeJSEnum.val, currentFilterResults);
            });
            selectAllResourcesUIDButton.gameObject.SetActive(false);
            if (parentBrowser.isVREnabled) parentBrowser.ApplyHelp(selectAllResourcesUIDButton, "Select all resources from the the browser results (replacing existing selections).");
            else parentBrowser.ApplyHelp(selectAllResourcesUIDButton, "Select all resources from the the browser results (replacing existing selections). Hold the Ctrl Key whilst pressing 'Select All' to add all resources from the the browser results to the existing selections.");

            resourceCountTextJSS = new JSONStorableString("resourceCountText", "");
            resourceCountTextBox = UI.CreateBasicTextBox(parentBoxGO, resourceCountTextJSS, 90f, 20f, 22, resourceCountHOffset, resourceCountVOffset, AnchorPresets.bottomMiddle, true, 18);
            parentBrowser.ApplyHelp(resourceCountTextBox, "Total count of resources from the broswer filter results and the range displayed on the current page of results.");
        }
        public static void DeSelectColor(UIDynamicButton buttonUID)
        {
            if (buttonUID!=null) buttonUID.buttonColor = BrowserUI.lightGrey;
        }
        public static void SelectColor(UIDynamicButton buttonUID)
        {
            if (buttonUID!=null) buttonUID.buttonColor = Color.cyan;
        }
        public void RepopulateButtons(bool forceSelectedResourceRefresh=false)
        {         
            int buttonsPerPage = (int)Settings.maxButtonsPerPageJSF.val;
            int buttonCount = Math.Min(currentFilterResults.Count, Math.Min(buttonsPerPage, currentFilterResults.Count - ((currentFilterResultsPage - 1) * buttonsPerPage)));
            if ((currentFilterResultsPage - 1) * buttonsPerPage > currentFilterResults.Count) currentFilterResultsPage = 1;
            if (currentFilterResultsPage < 1) currentFilterResultsPage = 1;

            RefactorButtonRows(buttonCount, buttonsPerRow);

            foreach (var kvp in UIDButtonToRVGEDict) kvp.Value.CancelButtonImageUpdate();

            UIDButtonToRVGEDict.Clear();
            RVGEToUIDButtonDict.Clear();

            int resourceIndexOffset = (currentFilterResultsPage - 1) * buttonsPerPage;

            ResourceVersionGroupEntry firstSelectedRVGE = parentBrowser.resourceSelectionsUI.GetFirstSelectedRVGE(browserFiltersUI.resourceTypeJSEnum.val);
            if (firstSelectedRVGE == null || (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi && forceSelectedResourceRefresh && !currentFilterResults.Contains(firstSelectedRVGE)))
            {
                if (currentFilterResults.Count!=0) SelectRVGE(currentFilterResults[resourceIndexOffset], activeButtonUIDs[0], true);
            }


            if (currentFilterResultsPage == 1)
            {
                firstPageUIDButton.button.enabled = false;
                firstPageUIDButton.textColor = Color.white;
                prevPageUIDButton.button.enabled = false;
                prevPageUIDButton.textColor = Color.white;
            }
            else
            {
                firstPageUIDButton.button.enabled = true;
                firstPageUIDButton.textColor = Color.black;
                prevPageUIDButton.button.enabled = true;
                prevPageUIDButton.textColor = Color.black;
            }
            if (currentFilterResultsPage >= Math.Ceiling(currentFilterResults.Count / (float)buttonsPerPage))
            {
                nextPageUIDButton.button.enabled = false;
                nextPageUIDButton.textColor = Color.white;
            }
            else
            {
                nextPageUIDButton.button.enabled = true;
                nextPageUIDButton.textColor = Color.black;
            }

            if (resourceButtonSize == 0)
            {
                decResourceButtonSizesUIDButton.button.enabled = false;
                decResourceButtonSizesUIDButton.textColor = Color.white;
            }
            else {
                decResourceButtonSizesUIDButton.button.enabled = true;
                decResourceButtonSizesUIDButton.textColor = Color.black;
            }
            if (resourceButtonSize == 6)
            {
                incResourceButtonSizesUIDButton.button.enabled = false;
                incResourceButtonSizesUIDButton.textColor = Color.white;
            }
            else
            {
                incResourceButtonSizesUIDButton.button.enabled = true;
                incResourceButtonSizesUIDButton.textColor = Color.black;
            }

            resourceCountTextJSS.val = (resourceIndexOffset + 1).ToString() + " - " + (resourceIndexOffset + buttonCount).ToString() + " of " + currentFilterResults.Count.ToString();

            //           UIDButtonToRVGEDict.Clear();

            currentPageRVGEs.Clear();
            currentPageButtons.Clear();

            parentBrowser.resourceSelectionsUI.ClearSelectedRVGEButtons(browserFiltersUI.resourceTypeJSEnum.val);

            for (int buttonId = buttonCount - 1; buttonId >= 0; buttonId--)
            {
                ResourceVersionGroupEntry rvge = currentFilterResults[buttonId + resourceIndexOffset];
                currentPageRVGEs.Insert(0,rvge);

                UIDynamicButton buttonUID = activeButtonUIDs[buttonId];
                currentPageButtons.Insert(0,buttonUID);

                UIDButtonToRVGEDict.Add(buttonUID, rvge);
                RVGEToUIDButtonDict.Add(rvge, buttonUID);
                if (parentBrowser.resourceSelectionsUI.IsSelected(rvge, browserFiltersUI.selectModeJSEnum.val, buttonUID)) buttonUID.buttonColor = Color.cyan;
                else DeSelectColor(buttonUID);

                buttonUID.label = rvge.resourceDisplayName;

                buttonUID.button.gameObject.SetActive(true);

                buttonUID.button.onClick.RemoveAllListeners();
                buttonUID.button.onClick.AddListener(delegate {
                    SelectRVGE(rvge, buttonUID);
                });

                currentFilterResults[buttonId + resourceIndexOffset].LoadSprite(activeButtonUIDs[buttonId]);
            }

            for (int i = buttonCount; i < (activeResourceRowGOs.Count) * buttonsPerRow; i++)
            {
                activeButtonUIDs[i].label = "";
                activeButtonUIDs[i].button.gameObject.SetActive(false);
            }


        }

        public UIDynamicButton GetUIDButton(ResourceVersionGroupEntry rvge)
        {
            if (RVGEToUIDButtonDict.ContainsKey(rvge)) return RVGEToUIDButtonDict[rvge];
            return null;
        }

        public void RefactorButtonRows(int requiredButtonCount, int requiredButtonsPerRow)
        {
            int rowsRequired = (int)Math.Ceiling((double)requiredButtonCount / requiredButtonsPerRow);
            GameObject buttonListContentGO = parentBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;

            int newRowsRequired = rowsRequired - activeResourceRowGOs.Count;

            if (buttonUIDsByRow.Count > 0)
            {
                int startingButtonsPerRow = buttonUIDsByRow.First().Value.Count();

                // Change in number of buttons per row
                if (startingButtonsPerRow != requiredButtonsPerRow)
                {
                    // Decrease in buttons perRow
                    if (startingButtonsPerRow > requiredButtonsPerRow)
                    {
                        // Remove spare buttons from every row (active or not)
                        foreach (var kvp in buttonUIDsByRow)
                        {
                            // Remove the excess buttons and add them to the unparented buttons
                            for (int i = requiredButtonsPerRow; i < startingButtonsPerRow; i++)
                            {
                                UIDynamicButton button = kvp.Value.Last();
                                kvp.Value.Remove(button);
                                unparentedButtonUIDs.Add(button);
                                button.gameObject.SetActive(false);
                                UIDButtonToRVGEDict.Remove(button);
                                activeButtonUIDs.Remove(button);
                            }
                        }
                    }
                    // Increase in buttons per Row
                    else if (startingButtonsPerRow < requiredButtonsPerRow)
                    {
                        HashSet<GameObject> recycledRows = new HashSet<GameObject>();
                        // Cycle through every row
                        foreach (var kvp in buttonUIDsByRow)
                        {
                            // Check row is not tagged for recycling
                            if (!recycledRows.Contains(kvp.Key))
                            {
                                // Cycle through missing button
                                for (int i = startingButtonsPerRow; i < requiredButtonsPerRow; i++)
                                {
                                    // Check if there are no spare unparented buttons available
                                    if (unparentedButtonUIDs.Count == 0)
                                    {
                                        // Attempt to recycle an inactiveRow if it exists
                                        if (inactiveResourceRowGOs.Count > 0)
                                        {
                                            var rowGO = inactiveResourceRowGOs.First();

                                            // Remove all buttons from the inactive row
                                            // and add them to the unparented buttons
                                            var buttonsList = buttonUIDsByRow[rowGO];
                                            foreach (var button in buttonsList)
                                            {
                                                unparentedButtonUIDs.Add(button);
                                                UIDButtonToRVGEDict.Remove(button);
                                            }
                                            buttonsList.Clear();
                                            recycledRows.Add(rowGO);
                                            inactiveResourceRowGOs.Remove(rowGO);

                                        }
                                        // If rows are due to be removed, then recycle a row
                                        // and add the buttons to the unparented buttons
                                        else if (newRowsRequired < 0)
                                        {
                                            GameObject rowGO = activeResourceRowGOs.Last();

                                            activeResourceRowGOs.Remove(rowGO);
                                            var buttonsList = buttonUIDsByRow[rowGO];
                                            foreach (UIDynamicButton button in buttonsList)
                                            {
                                                activeButtonUIDs.Remove(button);
                                                unparentedButtonUIDs.Add(button);
                                                UIDButtonToRVGEDict.Remove(button);
                                                button.gameObject.SetActive(false);
                                            }
                                            buttonsList.Clear();
                                            recycledRows.Add(rowGO);
                                            rowGO.SetActive(false);
                                            newRowsRequired++;
                                        }
                                    }
                                    // If unparentedButtons exist then take one and add it to the row
                                    if (unparentedButtonUIDs.Count > 0)
                                    {
                                        UIDynamicButton button = unparentedButtonUIDs.Last();
                                        unparentedButtonUIDs.Remove(button);
                                        kvp.Value.Add(button);
                                        button.gameObject.SetActive(true);
                                    }
                                    else
                                    {
                                        kvp.Value.Add(CreateResourceButtonInRow(kvp.Key, i));
                                    }
                                }
                            }
                        }
                        foreach (var rowGO in recycledRows) buttonUIDsByRow.Remove(rowGO);
                        recycledRows.Clear();
                    }

                    activeButtonUIDs.Clear();
                    foreach (var rowGO in activeResourceRowGOs) ResizeResourceButtonRow(rowGO, true);
                    foreach (var rowGO in inactiveResourceRowGOs) ResizeResourceButtonRow(rowGO);
                }
            }

            if (newRowsRequired > 0)
            {
                for (int i = 0; i < newRowsRequired; i++)
                {
                    if (inactiveResourceRowGOs.Count == 0)
                    {
                        GameObject rowGO = UI.AddChildGOImage(buttonListContentGO, new Color(0f, 0.8f, 0f,0f), AnchorPresets.hStretchTop, parentBrowser.browserScreenWidth-20f, resouceButtonWidth + resourceButtonlabelHeight + 2, Vector2.zero);
                        activeResourceRowGOs.Add(rowGO);
                        List<UIDynamicButton> buttonsInRow = new List<UIDynamicButton>();
                        buttonUIDsByRow.Add(rowGO, buttonsInRow);
                        for (int column = 0; column < requiredButtonsPerRow; column++)
                        {
                            UIDynamicButton resourceButton = CreateResourceButtonInRow(rowGO, column);
                            buttonsInRow.Add(resourceButton);
                            activeButtonUIDs.Add(resourceButton);
                        }
                    }
                    else
                    {
                        GameObject rowGO = inactiveResourceRowGOs.Last();
                        inactiveResourceRowGOs.Remove(rowGO);
                        rowGO.SetActive(true);
                        activeResourceRowGOs.Add(rowGO);
                        foreach (UIDynamicButton button in buttonUIDsByRow[rowGO])
                        {
                            activeButtonUIDs.Add(button);
                        }
                        
                    }

                }
            }
            else if (newRowsRequired < 0)
            {
                for (int i = 0; i < Math.Abs(newRowsRequired); i++)
                {
                    GameObject rowGO = activeResourceRowGOs.Last();

                    activeResourceRowGOs.Remove(rowGO);
                    foreach (UIDynamicButton button in buttonUIDsByRow[rowGO]) activeButtonUIDs.Remove(button);
                    inactiveResourceRowGOs.Add(rowGO);
                    rowGO.SetActive(false);
                }
            }
        }
        private void ResizeResourceButtonRow(GameObject rowGO, bool addtoActiveButtons = false)
        {
            UI.ResizeHeight(rowGO, resouceButtonWidth + resourceButtonlabelHeight + 2);
            int column = 0;
            foreach (var button in buttonUIDsByRow[rowGO])
            {
                button.transform.SetParent(rowGO.transform);
                button.buttonImage.enabled = true;
                button.button.enabled = true;

                UI.ResizeRecTransform(button.gameObject, resouceButtonWidth, resouceButtonWidth + resourceButtonlabelHeight, (column * resouceButtonWidth) + 2 + (column * 3), 0, AnchorPresets.middleLeft);

                Transform textTransform = button.button.transform.Find("Text");
                UI.ResizeRecTransform(textTransform.gameObject, resouceButtonWidth * 2, (resourceButtonlabelHeight - 1) * 2, 0, 0, AnchorPresets.bottomLeft);

                Transform imageTransform = button.button.transform.Find("ButtonSubImage");
                UI.ResizeRecTransform(imageTransform.gameObject, resouceButtonWidth, resouceButtonWidth, 0, -2f, AnchorPresets.topMiddle);

                if (addtoActiveButtons) activeButtonUIDs.Add(button);

                column++;
            }
        }
        private UIDynamicButton CreateResourceButtonInRow(GameObject rowGO, int column)
        {
            UIDynamicButton button = UI.CreateUIButton(rowGO, resouceButtonWidth, resouceButtonWidth + resourceButtonlabelHeight, "", 20, (column * resouceButtonWidth) + 2 + (column * 3), 0, AnchorPresets.middleLeft);

            GameObject newImageGO = UI.AddChildGOImage(button.button.gameObject, Color.white, AnchorPresets.topMiddle, button.height - 18f, button.height - 18f, new Vector2(0f, -2f));
            newImageGO.name = "ButtonSubImage";

            Transform textTransform = button.button.transform.Find("Text");
            UI.ResizeRecTransform(textTransform.GetComponent<RectTransform>(), resouceButtonWidth * 2, (resourceButtonlabelHeight - 1) * 2, 0, 0, AnchorPresets.bottomLeft);
            button.buttonText.color = Color.black;
            button.buttonImage.sprite = null;
            button.buttonText.resizeTextForBestFit = true;
            button.buttonText.resizeTextMinSize = 15;
            button.buttonText.resizeTextMaxSize = 21;

            //            activeButtonUIDs.Add(button);

            ResourceButtonPointerEnterExit click = button.button.gameObject.AddComponent<ResourceButtonPointerEnterExit>();
            click.Init(button, UIDButtonToRVGEDict, parentBrowser.helpTextJSS);

            return button;
        }

        private bool tempMultiMode =false;
        private int previousSelecteMode;

        private bool multiSelectKeyPressed { get
            {
                return Input.GetKey(KeyCode.LeftShift) || Input.GetKey(KeyCode.RightShift) || Input.GetKey(KeyCode.LeftAlt) || Input.GetKey(KeyCode.RightAlt) || Input.GetKey(KeyCode.LeftControl) || Input.GetKey(KeyCode.RightControl);
            } }
        public void Update()
        {
            if (!parentBrowser.isVREnabled)
            {                
                if (multiSelectKeyPressed)
                {
                    if (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi)
                    {
                        previousSelecteMode = browserFiltersUI.selectModeJSEnum.val;
                        browserFiltersUI.selectModeJSEnum.val = BrowserSelectMode.multi;
                        tempMultiMode = true;
                    }
                    
                }
                else if (tempMultiMode)
                {
                    browserFiltersUI.selectModeJSEnum.val = previousSelecteMode;
                    tempMultiMode = false;
                }
            }

        }

        private void SelectRVGE(ResourceVersionGroupEntry rvge, UIDynamicButton buttonUID, bool suppressQuickActivate=false)
        {
            if (browserFiltersUI.selectModeJSEnum.val != BrowserSelectMode.multi)
            {
                parentBrowser.resourceSelectionsUI.SingleSelectRVGE(rvge, buttonUID);
                if (browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.instant && !suppressQuickActivate) rvge.Activate(parentBrowser.resourceActionsUI.primaryActionUITab);
            }
            else if (!parentBrowser.isVREnabled)
            {
                if (multiSelectKeyPressed) tempMultiMode = false;
                if (Input.GetKey(KeyCode.LeftShift)) parentBrowser.resourceSelectionsUI.LeftShiftSelectRVGE(rvge, buttonUID, currentPageRVGEs, currentPageButtons);
                else if (Input.GetKey(KeyCode.RightShift)) parentBrowser.resourceSelectionsUI.RightShiftSelectRVGE(rvge, buttonUID, currentPageRVGEs, currentPageButtons);
                else if (Input.GetKey(KeyCode.LeftControl) || Input.GetKey(KeyCode.RightControl)) parentBrowser.resourceSelectionsUI.ToggleSelectRVGE(rvge, buttonUID);
                else if (Input.GetKey(KeyCode.LeftAlt) || Input.GetKey(KeyCode.RightAlt)) parentBrowser.resourceSelectionsUI.SingleSelectRVGE(rvge, buttonUID);
                else
                {
                    parentBrowser.resourceSelectionsUI.DeSelectAll(rvge.resourceType);
                    parentBrowser.resourceSelectionsUI.SingleSelectRVGE(rvge, buttonUID);
                }
            }
            else parentBrowser.resourceSelectionsUI.ToggleSelectRVGE(rvge, buttonUID);
        }

        public void RepositionNavButtons()
        {
            UI.ResetAnchorPosition(firstPageUIDButton.gameObject, navButtonsHOffset, navButtonsVOffset);
            UI.ResetAnchorPosition(prevPageUIDButton.gameObject, navButtonsHOffset + 15, navButtonsVOffset);
            UI.ResetAnchorPosition(nextPageUIDButton.gameObject, navButtonsHOffset + 30, navButtonsVOffset);

            UI.ResetAnchorPosition(incResourceButtonSizesUIDButton.gameObject, resoureButtonSizeHOffset , navButtonsVOffset);
            UI.ResetAnchorPosition(decResourceButtonSizesUIDButton.gameObject, resoureButtonSizeHOffset+20, navButtonsVOffset);

            UI.ResetAnchorPosition(resourceCountTextBox.gameObject, resourceCountHOffset, resourceCountVOffset);

            UI.ResetAnchorPosition(clearMultiSelectResourcesUIDButton.gameObject, navButtonsHOffset, navButtonsVOffset-17f);
            UI.ResetAnchorPosition(selectAllResourcesUIDButton.gameObject, navButtonsHOffset+53f, navButtonsVOffset - 17f);
            SetMultiSelectClear();
        }

        public void SetMultiSelectClear()
        {
          if (browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi && parentBrowser.resourceSelectionsUI.SelectionCount(browserFiltersUI.resourceTypeJSEnum.val) > 1) clearMultiSelectResourcesUIDButton.gameObject.SetActive(true);
           else clearMultiSelectResourcesUIDButton.gameObject.SetActive(false);

            if (browserFiltersUI.selectModeJSEnum.val == BrowserSelectMode.multi) selectAllResourcesUIDButton.gameObject.SetActive(true);
            else selectAllResourcesUIDButton.gameObject.SetActive(false);
        }
    }
    
    
    public class BrowserFiltersUI
    {
        public JSONStorableMultiEnumStringChooser creatorJSMultiEnum;
        public JSONStorableEnumStringChooser resourceTypeJSEnum;
        public JSONStorableStringChooser folderFilterJSSC;
        public JSONStorableString resourceNameFilterJSS;

        public JSONStorableBool showHiddenJSB;
        public JSONStorableBool onlyFavouritesJSB;

        public JSONStorableMultiEnumStringChooser tagsFilterJSMultiEnum;
        public JSONStorableEnumStringChooser licenseFilterJSEnum;

        public JSONStorableEnumStringChooser selectModeJSEnum;
        public JSONStorableEnumStringChooser sortModeJSEnum;

        public string resourceNameFilterQuick;

        private UIDynamicButton refreshButtonUID;
        private UIDynamicPopup folderPopupUID;
        private UIDynamicPopup licensePopupUID;
        private UIDynamicButton clearResourceNameFilterButton;

        private BrowserUI parentBrowser;
        public BrowserFiltersUI(BrowserUI _parentBrowser)
        {
            parentBrowser = _parentBrowser;
            resourceTypeJSEnum = new JSONStorableEnumStringChooser("resourceType", ResourceType.enumManifestName, ResourceType.scene, "Resource Type", ResourceTypeFilterCallback);

            List<int> resourceTypesToExclude = new List<int>();
            if (!PC.patreonContentEnabled)
            {
                resourceTypesToExclude = EnumManifest.GetEnumsList(ResourceType.enumManifestName);
                resourceTypesToExclude.Remove(ResourceType.scene);
                resourceTypesToExclude.Remove(ResourceType.presetAppearance);
            }
            resourceTypesToExclude.Add(ResourceType.plugins);

            resourceTypeJSEnum.SetEnumChoices(ResourceType.enumManifestName, resourceTypesToExclude);

             creatorJSMultiEnum = new JSONStorableMultiEnumStringChooser("creatorFilter", null, null, null,null, CreatorsMainValCallback, CreatorFilterChoices.enumManifestName,CreatorsTopValCallback);
            creatorJSMultiEnum.label = "Creator";

            folderFilterJSSC = new JSONStorableStringChooser("folderFilter", null, "", "Folder");
            resourceNameFilterJSS = new JSONStorableString("resourceNameFilter", "");
            resourceNameFilterQuick = "";

            showHiddenJSB = new JSONStorableBool("showHidden", false,ShowHiddenFilterCallback);
            onlyFavouritesJSB = new JSONStorableBool("onlyFavourites", false,OnlyFavFilterCallback);

            tagsFilterJSMultiEnum = new JSONStorableMultiEnumStringChooser("tagsFilter", TagManifest.GetTagNames(resourceTypeJSEnum.val), TagManifest.GetTagNames(resourceTypeJSEnum.val),null,null,TagMainValCallback, TagGroups.enumManifestName, TagTopValCallback);
            tagsFilterJSMultiEnum.label = "Tags";

            licenseFilterJSEnum = new JSONStorableEnumStringChooser("licenseFilter",LicenseType.enumManifestName,LicenseType.Any,"Package License",LicenseCallback);

            selectModeJSEnum = new JSONStorableEnumStringChooser ("selectMode",BrowserSelectMode.enumManifestName,BrowserSelectMode.single,"Select Mode",SelectModeCallback);
            sortModeJSEnum = new JSONStorableEnumStringChooser("sortMode", BrowserSortMode.enumManifestName, BrowserSortMode.cataloguedDateTime, "Order by",SortByCallback);

            creatorJSMultiEnum.valTopEnumNoCallback = CreatorFilterChoices.allVARs;
        }

        

        public void InitUI(GameObject filtersBoxGO)
        {
            UIDynamicPopup resourceType = UI.CreatePopup(filtersBoxGO, resourceTypeJSEnum, 160f, 30f, "", 20, 0f, 0f, AnchorPresets.topLeft);
            parentBrowser.ApplyHelp(resourceType, "Select a type of VAM resource to apply as a filter to the Browser results.");

            UIDynamicPopup tagPopupUID = UI.CreateFilterablePopup(filtersBoxGO, tagsFilterJSMultiEnum, 160f, 30f, "", 20, 165f, 0f, 150f, AnchorPresets.topLeft);  
            tagPopupUID.popup.onOpenPopupHandlers += RefreshTagFilters;
            parentBrowser.ApplyHelp(tagPopupUID, "Select a Tag to apply as a filter to the Browser results.");


            UIDynamicPopup creatorPopupUID = UI.CreateFilterablePopup(filtersBoxGO, creatorJSMultiEnum, 160f, 30f, "", 20, 330f, 0f, 300f, AnchorPresets.topLeft);
            creatorPopupUID.popup.onOpenPopupHandlers += RefreshCreatorFilters;
            parentBrowser.ApplyHelp(creatorPopupUID, "Select a Creator to filter Browser results by a specific VAR package creator. Non VAR packaged resources can be selected with the 'Local' option.");

//            folderPopupUID = UI.CreatePopup(filtersBoxGO, folderFilterJSSC, 115f, 15f, "", 20, -35f, 0f, AnchorPresets.topRight);

            licensePopupUID = UI.CreatePopup(filtersBoxGO,licenseFilterJSEnum, 115f, 30f, "", 20, -35f, 0f, AnchorPresets.topRight);
            parentBrowser.ApplyHelp(licensePopupUID, "Select a License type to filter Browser results by the license type applicable to the containing VAR package.");

            refreshButtonUID = UI.CreateUIButton(filtersBoxGO, 33f, 30f, "Rescan for new VARs", 16, 0f, 0f, AnchorPresets.topRight);
            refreshButtonUID.button.image.sprite = null;
            refreshButtonUID.buttonColor = Color.blue;
            refreshButtonUID.textColor = Color.white;

            refreshButtonUID.button.onClick.AddListener(delegate
            {
                SuperController.singleton.StartCoroutine(RefreshContents(refreshButtonUID));
            });
            parentBrowser.ApplyHelp(refreshButtonUID, "Scans for new VARs in the AddonPackages folder. Also updates Hidden & Favourite states of resources changed in VAM UI.");


            UIDynamicPopup selectModePopup = UI.CreatePopup(filtersBoxGO, selectModeJSEnum, 160f, 20f, "", 20, 0f, -30f, AnchorPresets.topLeft);
            parentBrowser.ApplyHelp(selectModePopup, "Instant Mode: the default action (e.g. Load Scene) will be performed instantly when selecting a resource in the browser. Single Mode: only one resource is selectable at a time. Multi Mode: many resources can be selected and modified simultaneously.");

            UIDynamicToggle hiddenUID = UI.CreateToggle(filtersBoxGO, showHiddenJSB, 85f, 20f, "Show Hidden", 20, 165f, -30f, AnchorPresets.topLeft);
            parentBrowser.ApplyHelp(hiddenUID, "Show resources that are marked as Hidden in the resource browser.");
            UIDynamicToggle onlyFavUID = UI.CreateToggle(filtersBoxGO, onlyFavouritesJSB, 95f, 20f, "Only Favorites", 20, 250f, -30f, AnchorPresets.topLeft);
            parentBrowser.ApplyHelp(onlyFavUID, "Only show resources that are marked as Favourite in the resource browser.");

            UIDynamicTextField textUID = UI.CreateTextBox(filtersBoxGO, resourceNameFilterJSS, 150f, 20f, 20, 345f, -30f, AnchorPresets.topLeft, true, InputField.LineType.MultiLineNewline, "Search Resource...", vScrollBarWidth: 0f,textEntryChangeCallback: ResourceNameFilterCallback);
            parentBrowser.ApplyHelp(textUID, "Filter resources where the resouce name starts with the specified search text");
            clearResourceNameFilterButton = UI.CreateUIButton(filtersBoxGO, 16f, 16f, "X", 20, 477f, -32, AnchorPresets.topLeft);
            clearResourceNameFilterButton.button.onClick.AddListener(delegate { resourceNameFilterJSS.val = "";});
            parentBrowser.ApplyHelp(clearResourceNameFilterButton, "Clear the Resource Name search field of user entered text.");
            if (resourceNameFilterJSS.val == "") clearResourceNameFilterButton.gameObject.SetActive(false);


            UIDynamicPopup sortModePopup = UI.CreatePopup(filtersBoxGO, sortModeJSEnum, 145f, 20f, "", 20, 0f, -30f, AnchorPresets.topRight);
            parentBrowser.ApplyHelp(sortModePopup, "Order the Browser resource results by Alphabetical resource name, by Package and Folder or by Date that the resource was first catalogued by Browser Assist.");

            RefreshFilters();
        }

        private IEnumerator RefreshContents(UIDynamicButton refreshButtonUID)
        {
            string tempLabel = refreshButtonUID.label;
            Color buttonCol =refreshButtonUID.buttonColor;
            refreshButtonUID.label = "Scan\n-ing";
            refreshButtonUID.buttonColor = Color.cyan;
            refreshButtonUID.textColor = Color.white;

            yield return new WaitForSeconds(0.01f);

            DateTime scanTime = DateTime.Now;

            if ((creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag && creatorJSMultiEnum.valTopEnum != CreatorFilterChoices.localOnly) || creatorJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.topEnumValFlag)
            {
                SuperController.singleton.RescanPackages();
                HashSet<string> varFiles = Utils.GetFilesAtPathRecursive("AddonPackages", "*.var");
                VARPackageManifest.RefreshManifest(varFiles, scanTime);
                VARPackageManifest.RefreshVARUserFilePrefs();
            }

            if (creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag && (creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.localOnly || creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.all)) ResourceManifest.LocalResourcesInit(scanTime);

            parentBrowser.RefreshFilterResults(true);

            refreshButtonUID.buttonText.text = tempLabel;
            refreshButtonUID.buttonColor = Color.blue;
 //           refreshButtonUID.textColor = Color.black;
        }

        private void RefreshTagFilters()
        {
            List<string> tags = TagManifest.GetTagNames(resourceTypeJSEnum.val);
            tagsFilterJSMultiEnum.SetMainChoices(tags);
            if (tagsFilterJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag)
            {
                if (!tags.Contains(tagsFilterJSMultiEnum.mainVal)) tagsFilterJSMultiEnum.valTopEnum = TagGroups.anyTags;
            }

            List<int> ignoreTagGroups = new List<int>();
            var tagCategories = TagManifest.GetTagCategories(resourceTypeJSEnum.val).ToList();
            tagCategories.Remove(TagCategory.userDefined);
            if (tagCategories.Count == 0) ignoreTagGroups.Add(TagGroups.noUserTags);

            tagsFilterJSMultiEnum.SetTopEnum(TagGroups.enumManifestName, ignoreTagGroups);

            if (tagsFilterJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag)
            {
                if (ignoreTagGroups.Contains(tagsFilterJSMultiEnum.valTopEnum)) tagsFilterJSMultiEnum.valTopEnum = TagGroups.anyTags;
            }
        }
        private void RefreshCreatorFilters()
        {
            List<string> creators = CreatorManifest.GetFilteredCreatorNames(this);
            List<string> displayCreators = CreatorManifest.GetFilteredCreatorNamesWithResourceCount(this);


            creatorJSMultiEnum.SetMainChoices(creators, displayCreators, false);

            if (creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag)
            {
                if (!creators.Contains(creatorJSMultiEnum.mainVal)) creatorJSMultiEnum.valTopEnum = CreatorFilterChoices.allVARs;
            }
            RefreshLicenseFilters();
        }

        private void RefreshLicenseFilters()
        {
            if (licensePopupUID.gameObject.activeSelf)
            {
                licenseFilterJSEnum.SetEnumChoices(LicenseType.enumManifestName, ResourceManifest.GetLicenseTypesByInverseFilters(resourceTypeJSEnum.val, showHiddenJSB.val, onlyFavouritesJSB.val, tagsFilterJSMultiEnum, creatorJSMultiEnum));
            }
        }
        private void RefreshFilters()
        {
            RefreshTagFilters();
            RefreshCreatorFilters();
            RefreshLicenseFilters();
        }
        private void CreatorsTopValCallback(int i)
        {
            licensePopupUID.gameObject.SetActive(true);
            if (i == CreatorFilterChoices.allVARs)
            {
                refreshButtonUID.label = "Rescan for new VARs";
                parentBrowser.ApplyHelp(refreshButtonUID, "Scans for new VARs in the AddonPackages folder. Also updates Hidden & Favourite states of resources changed in VAM UI.");
            }
            else
            {
                if (i == CreatorFilterChoices.all)
                {
                    refreshButtonUID.label = "Rescan Local & VARs";
                    parentBrowser.ApplyHelp(refreshButtonUID, "Scans for new VARs in the AddonPackages folder and new (or removed) files in the local Saves and Custom folders. Also updates Hidden & Favourite states of resources changed in VAM UI.");
                }
                else
                {
                    refreshButtonUID.label = "Rescan Local files";
                    parentBrowser.ApplyHelp(refreshButtonUID, "Scans for new (or removed) files in the local Saves and Custom folders. Also updates Hidden & Favourite states of resources changed in VAM UI.");
                }
                licensePopupUID.gameObject.SetActive(false);
            }
            RefreshLicenseFilters();
            parentBrowser.RefreshFilterResults(true);
        }
        private void CreatorsMainValCallback(string i)
        {
            refreshButtonUID.label = "Rescan for new VARs";
            parentBrowser.ApplyHelp(refreshButtonUID, "Scans for new VARs in the AddonPackages folder. Also updates Hidden & Favourite states of resources changed in VAM UI.");
            RefreshLicenseFilters();
            parentBrowser.RefreshFilterResults(true);
            licensePopupUID.gameObject.SetActive(true);
        }

        private void TagTopValCallback(int i)
        {
            if (i == TagGroups.hiddenOnly)
            {
                showHiddenJSB.valNoCallback = true;
                ShowHiddenFilterCallback(true);
                return;
            }
            RefreshCreatorFilters();
            parentBrowser.RefreshFilterResults(true);
           
        }
        private void TagMainValCallback(string i)
        {
            RefreshCreatorFilters();
            parentBrowser.RefreshFilterResults(true);
            
        }

        private void ResourceNameFilterCallback(string i)
        {
            resourceNameFilterQuick = i.ToLower();
            if (resourceNameFilterQuick == "") clearResourceNameFilterButton.gameObject.SetActive(false);
            else clearResourceNameFilterButton.gameObject.SetActive(true);
            parentBrowser.RefreshFilterResults(true);
            
        }

        private void ResourceTypeFilterCallback(string value)
        {
            RefreshFilters();
            parentBrowser.RefreshFilterResults(true);
            
            parentBrowser.resourceActionsUI.RefreshUI();
            parentBrowser.resourceSelectionsUI.RePopulateSelectionsBox();
        }
        private void ShowHiddenFilterCallback(bool value)
        {
//            if (!value) VARPackageManifest.RefreshVARUserFilePrefs();
            RefreshCreatorFilters();
            parentBrowser.RefreshFilterResults(true);
            parentBrowser.resourceActionsUI.RefreshUI();
        }
        private void OnlyFavFilterCallback(bool value)
        {
//            if (value) VARPackageManifest.RefreshVARUserFilePrefs();
            RefreshCreatorFilters();
            parentBrowser.RefreshFilterResults(true);
            parentBrowser.resourceActionsUI.RefreshUI();
        }
        private void SortByCallback(string i)
        {
            parentBrowser.RefreshFilterResults();
        }
        private void SelectModeCallback(int i)
        {
            parentBrowser.resourceSelectionsUI.SwitchSelectMode(resourceTypeJSEnum.val,i);
            parentBrowser.resourceActionsUI.RefreshUI();
            parentBrowser.RefreshBrowserUILayout();
        }

        private void LicenseCallback(int i)
        {
            parentBrowser.RefreshFilterResults();
        }
    }


}
