using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.UI;
using MeshVR;

namespace JayJayWon
{

    public static class Globals
    {
        public static MVRScript mvrScript;
        public static bool customFolderAccess = false;
        public static bool savesFolderAccess = false;
        public static bool addonPackageFilePrefsFolderAccess = false;

    }
    static class Const
    {
        public const float _localCanvasScale = 0.003f;

        public const string VARMetaDataFolder = "Saves\\PluginData\\JayJayWon\\BrowserAssist\\VARMetaData";
        public const string VARUserDataFolder = "Saves\\PluginData\\JayJayWon\\BrowserAssist\\VARUserData";
        public const string LocalUserDataFolder = "Saves\\PluginData\\JayJayWon\\BrowserAssist\\LocalUserData";
        public const string CreatorUserDataFolder = "Saves\\PluginData\\JayJayWon\\BrowserAssist\\CreatorUserData";
        public const string SymLinksFolder = "Saves\\PluginData\\JayJayWon\\BrowserAssist\\SymLinks";
        public const string  BAScriptFolder = "Custom\\Scripts\\JayJayWon\\BrowserAssist";
        public const string BAScriptDataFolder = "Saves\\PluginData\\JayJayWon\\BrowserAssist";
        public const string BAversion = "1.1.0";
        public const int BAMajorVersion = 1;
        public const int BAMinorVersion = 0;
        public const int BAFixVersion = 0;

        public const int VARManifestStoreFormat = 2;
        public const int VARUserDataStoreFormat = 2;
        public const int LocalUserDataStoreFormat = 2;

        public const string CreatorMetaDataFileExt = ".creatorManifest";
        public const string MetaDataFileExt = ".manifest";
        public const string UserDataFileExt = ".userData";
    }


    static class Utils
    {

        static public bool CreatePluginDataFolder()
        {
            bool result = false;
            if (FileManagerSecure.DirectoryExists("Saves\\PluginData"))
            {
                if (!FileManagerSecure.DirectoryExists(Const.VARMetaDataFolder))FileManagerSecure.CreateDirectory(Const.VARMetaDataFolder);
                if (!FileManagerSecure.DirectoryExists(Const.VARUserDataFolder)) FileManagerSecure.CreateDirectory(Const.VARUserDataFolder);
                if (!FileManagerSecure.DirectoryExists(Const.LocalUserDataFolder)) FileManagerSecure.CreateDirectory(Const.LocalUserDataFolder);
                if (!FileManagerSecure.DirectoryExists(Const.CreatorUserDataFolder)) FileManagerSecure.CreateDirectory(Const.CreatorUserDataFolder);
                if (FileManagerSecure.DirectoryExists(Const.VARMetaDataFolder)) result = true;
            }
            else SuperController.LogMessage("BrowserAssist: this function requires the folder Saves\\PluginData to exist.");
            return (result);
        }
        static public HashSet<string> GetFilesAtPathRecursive(string path, string pattern, bool ignoreDisabledFiles = false)
        {
            string[] patterns = new string[1];
            patterns[0] = pattern;
            return GetFilesAtPathRecursive(path, patterns, ignoreDisabledFiles);
        }

        static public HashSet<string> GetFilesAtPathRecursive(string path, string[] patterns, bool ignoreDisabledFiles=false)
        {
            IEnumerable<string> files;

            if (patterns.Contains("*.*")) files = FileManagerSecure.GetFiles(path);
            else files = FileManagerSecure.GetFiles(path).Where(fn => patterns.Contains("*."+fn.Substring(fn.LastIndexOf('.')+1)));

            string[] directories = FileManagerSecure.GetDirectories(path);
            HashSet<string> combined = new HashSet<string>(files);

            if (ignoreDisabledFiles)
            {

                HashSet<string> disabledVARFiles = new HashSet<string>();
                foreach (string file in combined)
                {
                    if (FileManagerSecure.FileExists(file + ".disabled"))
                    {
                        disabledVARFiles.Add(file);
                    }
                }
                combined.ExceptWith(disabledVARFiles);
            }
            

            directories.ToList().ForEach(directory =>
            {
                if (!directory.EndsWith(".var:")) combined.UnionWith(GetFilesAtPathRecursive(directory, patterns, ignoreDisabledFiles));

            });

            return combined;
        }

        public static string GetLatestVARPath(string path)
        {
            if (!path.Contains(".latest:") && path.Contains(":") && path.IndexOf(':') > 1)
            {
                string varPackageName = path.Substring(0, path.IndexOf(':'));
                string varFileName = path.Substring(varPackageName.Length + 1);
                path = varPackageName.Substring(0, varPackageName.LastIndexOf('.')) + ".latest:" + varFileName;
            }
            path = SuperController.singleton.NormalizeLoadPath(path);

            return (path);
        }
        static public void LogDebug(string message)
        {
            SuperController.LogMessage("BrowserAssist: " + message);
        }

    }

    public static class ImageUtils
    {
        public static ImageLoaderThreaded.QueuedImage QueueLoadTexture(string url, ImageLoaderThreaded.ImageLoaderCallback callback, bool immediate)
        {
            if (ImageLoaderThreaded.singleton == null)
                return null;

            ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
            queuedImage.imgPath = url;
            queuedImage.forceReload = true;
            queuedImage.skipCache = true;
            queuedImage.compress = true;
            queuedImage.createMipMaps = false;
            queuedImage.isNormalMap = false;
            queuedImage.linear = false;
            queuedImage.createAlphaFromGrayscale = false;
            queuedImage.createNormalFromBump = false;
            queuedImage.bumpStrength = 0f;
            queuedImage.isThumbnail = false;
            queuedImage.fillBackground = false;
            queuedImage.invert = false;
            queuedImage.callback = callback;

            Texture2D tex = ImageLoaderThreaded.singleton.GetCachedThumbnail(url);
            if (tex != null)
            {
                queuedImage.tex = tex;
                callback.Invoke(queuedImage);
            }
            else
            {
                if (immediate) ImageLoaderThreaded.singleton.QueueThumbnailImmediate(queuedImage);
                else ImageLoaderThreaded.singleton.QueueThumbnail(queuedImage);
            }

            return queuedImage;
        }
    }

    public static class AtomUtils
    {
        public static List<string> GetAtomNames(bool includeOffAtoms = false, bool includeCoreControl = false, bool includeCameraRig = false, bool includePersonAtoms = true, bool incudeNonPersonAtoms = true)
        {
            List<Atom> atoms = GetAtoms(includeOffAtoms, includeCoreControl, includeCameraRig, includePersonAtoms, incudeNonPersonAtoms);
            List<string> atomNames = new List<string>();
            foreach (Atom atom in atoms) atomNames.Add(atom.name);
            return atomNames;
        }
        public static List<Atom> GetAtoms(bool includeOffAtoms = false, bool includeCoreControl = false, bool includeCameraRig = false, bool includePersonAtoms = true, bool incudeNonPersonAtoms=true)
        {
            List<Atom> atomsList = new List<Atom>();
            foreach (Atom atom in SuperController.singleton.GetAtoms())
            {
                if ((atom.on && (atom.containingSubScene == null || atom.containingSubScene.containingAtom.on)) || includeOffAtoms)
                {
                    if (atom.name == "CoreControl")
                    {
                        if (includeCoreControl) atomsList.Add(atom);
                    }
                    else if (atom.name == "[CameraRig]")
                    {
                        if (includeCameraRig) atomsList.Add(atom);
                    }
                    else if (atom.type == "Person" && includePersonAtoms) atomsList.Add(atom);
                    else if (atom.type != "Person" && incudeNonPersonAtoms) atomsList.Add(atom);
                }
            }

            return (atomsList);
        }

        public static bool IsMale(Atom atom)
        {
            bool isMale = false;
            // If the peson atom is not "On", then we cant determine their gender it seems as GetComponentInChildren<DAZCharacter> just returns null
            if (atom.on && (atom.containingSubScene == null || atom.containingSubScene.containingAtom.on) && atom.type == "Person") isMale = atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("male") ||
                    atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarlee") ||
                    atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarjulian") ||
                    atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("Futa");
            return (isMale);
        }

        public static bool IsFemale(Atom atom)
        {
            bool isFemale = false;
            // If the peson atom is not "On", then we cant determine their gender it seems as GetComponentInChildren<DAZCharacter> just returns null
            if (atom.on && (atom.containingSubScene == null || atom.containingSubScene.containingAtom.on) && atom.type == "Person" && !IsMale(atom)) isFemale = true;
            return (isFemale);
        }

        public static string GetNextAvailableAtomName(string preferredAtomName)
        {
            List<Atom> atoms = GetAtoms(true, true, true);
            List<string> atomNames = new List<string>();
            foreach (Atom tempAtom in atoms) atomNames.Add(tempAtom.uid);

            if (atomNames.Contains(preferredAtomName))
            {
                for (int i = 2; i < 10000; i++)
                {
                    string newAtomName = preferredAtomName + "#" + i.ToString();
                    if (!atomNames.Contains(newAtomName)) return newAtomName;
                }
            }
            else return preferredAtomName;
            return "";
        }


    }

    public static class PluginUtils
    {

        // This fuction assumes that the plugin is a Session plugin
        public static List<MVRScript> GetSessionPlugins(string pluginTypeFilter = null)
        {
            List<MVRScript> pluginList = new List<MVRScript>();
            foreach (Transform transform in Globals.mvrScript.manager.gameObject.transform.Find("Plugins"))
            {
                MVRScript script = transform.gameObject.GetComponent<MVRScript>();
                if (script != null && (pluginTypeFilter == null || script.name.EndsWith(pluginTypeFilter))) pluginList.Add(script);
            }

            return (pluginList);
        }

        public static List<MVRScript> GetScenePlugins(string pluginTypeFilter = null)
        {
            List<MVRScript> pluginList = new List<MVRScript>();

            foreach (Atom atom in SuperController.singleton.GetAtoms())
            {
                MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
                if (manager != null)
                {
                    foreach (Transform transform in manager.gameObject.transform.Find("Plugins"))
                    {
                        MVRScript pluginStorable = transform.gameObject.GetComponent<MVRScript>();
                        if (pluginStorable != null && (pluginTypeFilter == null || pluginStorable.name.EndsWith(pluginTypeFilter))) pluginList.Add(pluginStorable);
                    }
                }
            }

            return (pluginList);
        }
        public static List<MVRScript> GetSceneAndSessionPlugins(string pluginTypeFilter = null)
        {
            return (GetSessionPlugins(pluginTypeFilter).Concat(GetScenePlugins(pluginTypeFilter)).ToList());
        }

        public static List<MVRScript> GetPluginsFromAtom(Atom atom, string pluginTypeFilter = null)
        {
            List<MVRScript> plugins = new List<MVRScript>();

            MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
            if (manager != null)
            {
                foreach (Transform transform in manager.gameObject.transform.Find("Plugins"))
                {
                    MVRScript pluginStorable = transform.gameObject.GetComponent<MVRScript>();
                    if (pluginStorable != null && (pluginTypeFilter == null || pluginStorable.name.EndsWith(pluginTypeFilter))) plugins.Add(pluginStorable);
                }
            }

            return plugins;
        }

        public static List<string> GetPluginRefsOfType(MVRPluginManager manager, string pluginPath)
        {
            List<string> pluginRefs = new List<string>();
            if (manager != null)
            {
                JSONClass currentPlugins = manager.GetJSON(true, true, true);
                if (currentPlugins != null)
                {
                    foreach (KeyValuePair<string, JSONNode> kvp in (JSONClass)currentPlugins["plugins"])
                    {
                        if (kvp.Value.ToString().Substring(kvp.Value.ToString().LastIndexOfAny(new char[] { '/', '\\' }) + 1) == pluginPath.Substring(pluginPath.LastIndexOfAny(new char[] { '/', '\\' }) + 1) + "\"") pluginRefs.Add(kvp.Key);
                    }
                }
            }
            return (pluginRefs);
        }
        public static List<string> GetPluginRefsOfType(Atom atom, string pluginPath)
        {
            MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;

            return GetPluginRefsOfType(manager, pluginPath);
        }

        public static List<string> GetSessionPluginRefsOfType(string pluginPath)
        {

            return GetPluginRefsOfType(Globals.mvrScript.manager, pluginPath);
        }

    }

    public class PresetLockStore
    {
        public bool _generalPresetLock;
        public bool _appPresetLock;
        public bool _posePresetLock;
        public bool _animationPresetLock;
        public bool _glutePhysPresetLock;
        public bool _breastPhysPresetLock;
        public bool _pluginPresetLock;
        public bool _skinPresetLock;
        public bool _morphPresetLock;
        public bool _hairPresetLock;
        public bool _clothingPresetLock;

        public void StorePresetLocks(Atom atom, bool clearAllLocks = false, bool lockClothingPreset = false)
        {
            List<PresetManagerControl> pmControlList = atom.presetManagerControls;
            foreach (PresetManagerControl pmc in pmControlList)
            {
                if (pmc.name == "geometry") _generalPresetLock = pmc.lockParams;
                if (pmc.name == "AppearancePresets") _appPresetLock = pmc.lockParams;
                if (pmc.name == "PosePresets") _posePresetLock = pmc.lockParams;
                if (pmc.name == "AnimationPresets") _animationPresetLock = pmc.lockParams;
                if (pmc.name == "FemaleGlutePhysicsPresets") _glutePhysPresetLock = pmc.lockParams;
                if (pmc.name == "FemaleBreastPhysicsPresets") _breastPhysPresetLock = pmc.lockParams;
                if (pmc.name == "PluginPresets") _pluginPresetLock = pmc.lockParams;
                if (pmc.name == "SkinPresets") _skinPresetLock = pmc.lockParams;
                if (pmc.name == "MorphPresets") _morphPresetLock = pmc.lockParams;
                if (pmc.name == "HairPresets") _hairPresetLock = pmc.lockParams;
                if (pmc.name == "ClothingPresets") _clothingPresetLock = pmc.lockParams;

                if (pmc.name == "ClothingPresets" && lockClothingPreset) pmc.lockParams = true;
                else if (clearAllLocks)
                {
                    pmc.lockParams = false;
                }
            }

        }
        public void RestorePresetLocks(Atom atom)
        {
            List<PresetManagerControl> pmControlList = atom.presetManagerControls;
            foreach (PresetManagerControl pmc in pmControlList)
            {
                if (pmc.name == "geometry") pmc.lockParams = _generalPresetLock;
                if (pmc.name == "AppearancePresets") pmc.lockParams = _appPresetLock;
                if (pmc.name == "PosePresets") pmc.lockParams = _posePresetLock;
                if (pmc.name == "AnimationPresets") pmc.lockParams = _animationPresetLock;
                if (pmc.name == "FemaleGlutePhysicsPresets") pmc.lockParams = _glutePhysPresetLock;
                if (pmc.name == "FemaleBreastPhysicsPresets") pmc.lockParams = _breastPhysPresetLock;
                if (pmc.name == "PluginPresets") pmc.lockParams = _pluginPresetLock;
                if (pmc.name == "SkinPresets") pmc.lockParams = _skinPresetLock;
                if (pmc.name == "MorphPresets") pmc.lockParams = _morphPresetLock;
                if (pmc.name == "HairPresets") pmc.lockParams = _hairPresetLock;
                if (pmc.name == "ClothingPresets") pmc.lockParams = _clothingPresetLock;
            }
        }
    }
}
